# Copyright 2023 The JaxGaussianProcesses Contributors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from dataclasses import dataclass

from beartype.typing import Mapping

from gpjax.dataset import Dataset
from gpjax.decision_making.utility_functions.base import (
    AbstractSinglePointUtilityFunctionBuilder,
    SinglePointUtilityFunction,
)
from gpjax.decision_making.utils import OBJECTIVE
from gpjax.gps import ConjugatePosterior
from gpjax.typing import KeyArray


@dataclass
class ThompsonSampling(AbstractSinglePointUtilityFunctionBuilder):
    """
    Form a utility function by drawing an approximate sample from the posterior,
    using decoupled sampling as introduced in [Wilson et. al.
    (2020)](https://arxiv.org/abs/2002.09309). Note that we return the *negative* of the
    sample as the utility function, as utility functions are *maximised*.

    Note that this is a single batch utility function, as it doesn't support classical
    batching. However, Thompson sampling can be used in a batched setting by drawing a
    batch of different samples from the GP posterior. This can be done by calling
    `build_utility_function` with different keys, an example of which can be seen in the
    `ask` method of the `UtilityDrivenDecisionMaker` class. The samples can then be
    optimised sequentially.

    Attributes:
        num_features (int): The number of random Fourier features to use when drawing
            the approximate sample from the posterior. Defaults to 100.
    """

    num_features: int = 100

    def __post_init__(self):
        if self.num_features <= 0:
            raise ValueError(
                "The number of random Fourier features must be a positive integer."
            )

    def build_utility_function(
        self,
        posteriors: Mapping[str, ConjugatePosterior],
        datasets: Mapping[str, Dataset],
        key: KeyArray,
    ) -> SinglePointUtilityFunction:
        """
        Draw an approximate sample from the posterior of the objective model and return
        the *negative* of this sample as a utility function, as utility functions
        are *maximised*.

        Args:
            posteriors (Mapping[str, AbstractPosterior]): Dictionary of posteriors to be
            used to form the utility function. One of the posteriors must correspond
            to the `OBJECTIVE` key, as we sample from the objective posterior to form
            the utility function.
            datasets (Mapping[str, Dataset]): Dictionary of datasets which may be used
            to form the utility function. Keys in `datasets` should correspond to
            keys in `posteriors`. One of the datasets must correspond
            to the `OBJECTIVE` key.
            key (KeyArray): JAX PRNG key used for random number generation. This can be
            changed to draw different samples.

        Returns:
            SinglePointUtilityFunction: An appproximate sample from the objective model
            posterior to to be *maximised* in order to decide which point to query
            next.
        """
        self.check_objective_present(posteriors, datasets)

        objective_posterior = posteriors[OBJECTIVE]
        if not isinstance(objective_posterior, ConjugatePosterior):
            raise ValueError(
                "Objective posterior must be a ConjugatePosterior to draw an approximate sample."
            )

        objective_dataset = datasets[OBJECTIVE]
        thompson_sample = objective_posterior.sample_approx(
            num_samples=1,
            train_data=objective_dataset,
            key=key,
            num_features=self.num_features,
        )

        return lambda x: -1.0 * thompson_sample(x)  # Utility functions are *maximised*



# Copyright 2023 The JaxGaussianProcesses Contributors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from beartype.typing import (
    Callable,
    Dict,
    Final,
)

from gpjax.dataset import Dataset
from gpjax.typing import (
    Array,
    Float,
)

OBJECTIVE: Final[str] = "OBJECTIVE"
"""
Tag for the objective dataset/function in standard utility functions.
"""


FunctionEvaluator = Callable[[Float[Array, "N D"]], Dict[str, Dataset]]
"""
Type alias for function evaluators, which take an array of points of shape $`[N, D]`$
and evaluate a set of functions at each point, returning a mapping from function tags
to datasets of the evaluated points. This is the same as the `Observer` in Trieste:
https://github.com/secondmind-labs/trieste/blob/develop/trieste/observer.py
"""


def build_function_evaluator(
    functions: Dict[str, Callable[[Float[Array, "N D"]], Float[Array, "N 1"]]]
) -> FunctionEvaluator:
    """
    Takes a dictionary of functions and returns a `FunctionEvaluator` which can be
    used to evaluate each of the functions at a supplied set of points and return a
    dictionary of datasets storing the evaluated points.
    """
    return lambda x: {tag: Dataset(x, f(x)) for tag, f in functions.items()}



from dataclasses import (
    dataclass,
    fields,
)
from functools import singledispatch

from beartype.typing import (
    Dict,
    Union,
)
from jaxlib.xla_extension import PjitFunction

from gpjax.decision_making.test_functions import (
    Forrester,
    LogarithmicGoldsteinPrice,
)
from gpjax.decision_making.utility_functions import ThompsonSampling
from gpjax.kernels import (
    RFF,
    ArcCosine,
    GraphKernel,
    Matern12,
    Matern32,
    Matern52,
)
from gpjax.objectives import (
    ELBO,
    CollapsedELBO,
    ConjugateMLL,
    LogPosteriorDensity,
    NonConjugateMLL,
)

CitationType = Union[None, str, Dict[str, str]]


@dataclass(repr=False)
class AbstractCitation:
    citation_key: Union[str, None] = None
    authors: Union[str, None] = None
    title: Union[str, None] = None
    year: Union[str, None] = None

    def as_str(self) -> str:
        citation_str = f"@{self.citation_type}{{{self.citation_key},"
        for field in fields(self):
            fn = field.name
            if fn not in ["citation_type", "citation_key", "notes"]:
                citation_str += f"\n{fn} = {{{getattr(self, fn)}}},"
        return citation_str + "\n}"

    def __repr__(self) -> str:
        return repr(self.as_str())

    def __str__(self) -> str:
        return self.as_str()


class NullCitation(AbstractCitation):
    def __str__(self) -> str:
        return (
            "No citation available. If you think this is an error, please open a pull"
            " request."
        )


@dataclass
class PhDThesisCitation(AbstractCitation):
    school: Union[str, None] = None
    institution: Union[str, None] = None
    citation_type: CitationType = "phdthesis"


@dataclass
class PaperCitation(AbstractCitation):
    booktitle: Union[str, None] = None
    citation_type: CitationType = "inproceedings"


@dataclass
class BookCitation(AbstractCitation):
    publisher: Union[str, None] = None
    volume: Union[str, None] = None
    citation_type: CitationType = "book"


####################
# Default citation
####################
@singledispatch
def cite(tree) -> AbstractCitation:
    return NullCitation()


####################
# Default citation
####################
@cite.register(PjitFunction)
def _(tree) -> None:
    raise RuntimeError("Citation not available for jitted objects.")


####################
# Kernel citations
####################
@cite.register(Matern12)
@cite.register(Matern32)
@cite.register(Matern52)
def _(tree) -> PhDThesisCitation:
    citation = PhDThesisCitation(
        citation_key="matern1960SpatialV",
        authors="Bertil Matérn",
        title=(
            "Spatial variation : Stochastic models and their application to some"
            " problems in forest surveys and other sampling investigations"
        ),
        year="1960",
        school="Stockholm University",
        institution="Stockholm University",
    )
    return citation


@cite.register(ArcCosine)
def _(_) -> PaperCitation:
    return PaperCitation(
        citation_key="cho2009kernel",
        authors="Cho, Youngmin and Saul, Lawrence",
        title="Kernel Methods for Deep Learning",
        year="2009",
        booktitle="Advances in Neural Information Processing Systems",
    )


@cite.register(GraphKernel)
def _(tree) -> PaperCitation:
    return PaperCitation(
        citation_key="borovitskiy2021matern",
        title="Matérn Gaussian Processes on Graphs",
        authors=(
            "Borovitskiy, Viacheslav and Azangulov, Iskander and Terenin, Alexander and"
            " Mostowsky, Peter and Deisenroth, Marc and Durrande, Nicolas"
        ),
        booktitle="International Conference on Artificial Intelligence and Statistics",
        year="2021",
    )


@cite.register(RFF)
def _(tree) -> PaperCitation:
    return PaperCitation(
        citation_key="rahimi2007random",
        authors="Rahimi, Ali and Recht, Benjamin",
        title="Random features for large-scale kernel machines",
        year="2007",
        booktitle="Advances in neural information processing systems",
        citation_type="article",
    )


####################
# Objective citations
####################
@cite.register(ConjugateMLL)
@cite.register(NonConjugateMLL)
@cite.register(LogPosteriorDensity)
def _(tree) -> BookCitation:
    return BookCitation(
        citation_key="rasmussen2006gaussian",
        title="Gaussian Processes for Machine Learning",
        authors="Rasmussen, Carl Edward and Williams, Christopher K",
        year="2006",
        publisher="MIT press Cambridge, MA",
        volume="2",
    )


@cite.register(CollapsedELBO)
def _(tree) -> PaperCitation:
    return PaperCitation(
        citation_key="titsias2009variational",
        title="Variational learning of inducing variables in sparse Gaussian processes",
        authors="Titsias, Michalis",
        year="2009",
        booktitle="International Conference on Artificial Intelligence and Statistics",
    )


@cite.register(ELBO)
def _(tree) -> PaperCitation:
    return PaperCitation(
        citation_key="hensman2013gaussian",
        title="Gaussian Processes for Big Data",
        authors="Hensman, James and Fusi, Nicolo and Lawrence, Neil D",
        year="2013",
        booktitle="Uncertainty in Artificial Intelligence",
        citation_type="article",
    )


####################
# Decision making citations
####################
@cite.register(ThompsonSampling)
def _(tree) -> PaperCitation:
    return PaperCitation(
        citation_key="wilson2020efficiently",
        title="Efficiently sampling functions from Gaussian process posteriors",
        authors="Wilson, James and Borovitskiy, Viacheslav and Terenin, Alexander and Mostowsky, Peter and Deisenroth, Marc",
        year="2020",
        booktitle="International Conference on Machine Learning",
        citation_type="article",
    )


@cite.register(Forrester)
def _(tree) -> BookCitation:
    return BookCitation(
        citation_key="forrester2008engineering",
        authors="Forrester, Alexander and Sobester, Andras and Keane, Andy",
        title="Engineering design via surrogate modelling: a practical guide",
        year="2008",
        publisher="John Wiley & Sons",
    )


@cite.register(LogarithmicGoldsteinPrice)
def _(tree) -> PaperCitation:
    return PaperCitation(
        citation_key="picheny2013benchmark",
        authors="Picheny, Victor and Wagner, Tobias and Ginsbourger, David",
        title="A benchmark of kriging-based infill criteria for noisy optimization",
        year="2013",
        booktitle="Structural and multidisciplinary optimization",
        citation_type="article",
    )



# Copyright 2022 The GPJax Contributors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import abc
from dataclasses import dataclass

from beartype.typing import (
    Any,
    Union,
)
import cola
from cola.linalg.decompositions.decompositions import Cholesky
import jax.numpy as jnp
import jax.scipy as jsp
from jaxtyping import Float
import tensorflow_probability.substrates.jax.bijectors as tfb

from gpjax.base import (
    Module,
    param_field,
    static_field,
)
from gpjax.dataset import Dataset
from gpjax.distributions import GaussianDistribution
from gpjax.gps import AbstractPosterior
from gpjax.likelihoods import Gaussian
from gpjax.lower_cholesky import lower_cholesky
from gpjax.typing import (
    Array,
    ScalarFloat,
)


@dataclass
class AbstractVariationalFamily(Module):
    r"""
    Abstract base class used to represent families of distributions that can be
    used within variational inference.
    """

    posterior: AbstractPosterior

    def __call__(self, *args: Any, **kwargs: Any) -> GaussianDistribution:
        r"""Evaluate the variational family's density.

        For a given set of parameters, compute the latent function's prediction
        under the variational approximation.

        Args:
            *args (Any): Arguments of the variational family's `predict` method.
            **kwargs (Any): Keyword arguments of the variational family's `predict`
                method.

        Returns
        -------
            GaussianDistribution: The output of the variational family's `predict` method.
        """
        return self.predict(*args, **kwargs)

    @abc.abstractmethod
    def predict(self, *args: Any, **kwargs: Any) -> GaussianDistribution:
        r"""Predict the GP's output given the input.

        Args:
            *args (Any): Arguments of the variational family's ``predict``
                method.
            **kwargs (Any): Keyword arguments of the variational family's
                ``predict`` method.

        Returns
        -------
            GaussianDistribution: The output of the variational family's ``predict`` method.
        """
        raise NotImplementedError


@dataclass
class AbstractVariationalGaussian(AbstractVariationalFamily):
    r"""The variational Gaussian family of probability distributions."""

    inducing_inputs: Float[Array, "N D"]
    jitter: ScalarFloat = static_field(1e-6)

    @property
    def num_inducing(self) -> int:
        """The number of inducing inputs."""
        return self.inducing_inputs.shape[0]


@dataclass
class VariationalGaussian(AbstractVariationalGaussian):
    r"""The variational Gaussian family of probability distributions.

    The variational family is $`q(f(\cdot)) = \int p(f(\cdot)\mid u) q(u) \mathrm{d}u`$, where
    $`u = f(z)`$ are the function values at the inducing inputs $`z`$
    and the distribution over the inducing inputs is
    $`q(u) = \mathcal{N}(\mu, S)`$.  We parameterise this over
    $`\mu`$ and $`sqrt`$ with $`S = sqrt sqrt^{\top}`$.
    """

    variational_mean: Union[Float[Array, "N 1"], None] = param_field(None)
    variational_root_covariance: Float[Array, "N N"] = param_field(
        None, bijector=tfb.FillTriangular()
    )

    def __post_init__(self) -> None:
        if self.variational_mean is None:
            self.variational_mean = jnp.zeros((self.num_inducing, 1))

        if self.variational_root_covariance is None:
            self.variational_root_covariance = jnp.eye(self.num_inducing)

    def prior_kl(self) -> ScalarFloat:
        r"""Compute the prior KL divergence.

        Compute the KL-divergence between our variational approximation and the
        Gaussian process prior.

        For this variational family, we have
        ```math
        \begin{align}
        \operatorname{KL}[q(f(\cdot))\mid\mid p(\cdot)] & = \operatorname{KL}[q(u)\mid\mid p(u)]\\
        & = \operatorname{KL}[ \mathcal{N}(\mu, S) \mid\mid N(\mu z, \mathbf{K}_{zz}) ],
        \end{align}
        ```
        where $`u = f(z)`$ and $`z`$ are the inducing inputs.

        Returns
        -------
             ScalarFloat: The KL-divergence between our variational
                approximation and the GP prior.
        """
        # Unpack variational parameters
        mu = self.variational_mean
        sqrt = self.variational_root_covariance
        z = self.inducing_inputs

        # Unpack mean function and kernel
        mean_function = self.posterior.prior.mean_function
        kernel = self.posterior.prior.kernel

        muz = mean_function(z)
        Kzz = kernel.gram(z)
        Kzz = cola.PSD(Kzz + cola.ops.I_like(Kzz) * self.jitter)

        sqrt = cola.ops.Triangular(sqrt)
        S = sqrt @ sqrt.T

        qu = GaussianDistribution(loc=jnp.atleast_1d(mu.squeeze()), scale=S)
        pu = GaussianDistribution(loc=jnp.atleast_1d(muz.squeeze()), scale=Kzz)

        return qu.kl_divergence(pu)

    def predict(self, test_inputs: Float[Array, "N D"]) -> GaussianDistribution:
        r"""Compute the predictive distribution of the GP at the test inputs t.

        This is the integral $`q(f(t)) = \int p(f(t)\mid u) q(u) \mathrm{d}u`$, which
        can be computed in closed form as:
        ```math
            \mathcal{N}\left(f(t); \mu t + \mathbf{K}_{tz} \mathbf{K}_{zz}^{-1} (\mu - \mu z),  \mathbf{K}_{tt} - \mathbf{K}_{tz} \mathbf{K}_{zz}^{-1} \mathbf{K}_{zt} + \mathbf{K}_{tz} \mathbf{K}_{zz}^{-1} S \mathbf{K}_{zz}^{-1} \mathbf{K}_{zt}\right).
        ```

        Args:
            test_inputs (Float[Array, "N D"]): The test inputs at which we wish to
                make a prediction.

        Returns
        -------
            GaussianDistribution: The predictive distribution of the low-rank GP at
                the test inputs.
        """
        # Unpack variational parameters
        mu = self.variational_mean
        sqrt = self.variational_root_covariance
        z = self.inducing_inputs

        # Unpack mean function and kernel
        mean_function = self.posterior.prior.mean_function
        kernel = self.posterior.prior.kernel

        Kzz = kernel.gram(z)
        Kzz += cola.ops.I_like(Kzz) * self.jitter
        Lz = lower_cholesky(Kzz)
        muz = mean_function(z)

        # Unpack test inputs
        t = test_inputs

        Ktt = kernel.gram(t)
        Kzt = kernel.cross_covariance(z, t)
        mut = mean_function(t)

        # Lz⁻¹ Kzt
        Lz_inv_Kzt = cola.solve(Lz, Kzt, Cholesky())

        # Kzz⁻¹ Kzt
        Kzz_inv_Kzt = cola.solve(Lz.T, Lz_inv_Kzt, Cholesky())

        # Ktz Kzz⁻¹ sqrt
        Ktz_Kzz_inv_sqrt = jnp.matmul(Kzz_inv_Kzt.T, sqrt)

        # μt + Ktz Kzz⁻¹ (μ - μz)
        mean = mut + jnp.matmul(Kzz_inv_Kzt.T, mu - muz)

        # Ktt - Ktz Kzz⁻¹ Kzt  +  Ktz Kzz⁻¹ S Kzz⁻¹ Kzt  [recall S = sqrt sqrtᵀ]
        covariance = (
            Ktt
            - jnp.matmul(Lz_inv_Kzt.T, Lz_inv_Kzt)
            + jnp.matmul(Ktz_Kzz_inv_sqrt, Ktz_Kzz_inv_sqrt.T)
        )
        covariance += cola.ops.I_like(covariance) * self.jitter

        return GaussianDistribution(
            loc=jnp.atleast_1d(mean.squeeze()), scale=covariance
        )


@dataclass
class WhitenedVariationalGaussian(VariationalGaussian):
    r"""The whitened variational Gaussian family of probability distributions.

    The variational family is $`q(f(\cdot)) = \int p(f(\cdot)\mid u) q(u) \mathrm{d}u`$,
    where $`u = f(z)`$
    are the function values at the inducing inputs $`z`$ and the distribution over
    the inducing inputs is $`q(u) = \mathcal{N}(Lz \mu + mz, Lz S Lz^{\top})`$. We parameterise this
    over $`\mu`$ and $`sqrt`$ with $`S = sqrt sqrt^{\top}`$.
    """

    def prior_kl(self) -> ScalarFloat:
        r"""Compute the KL-divergence between our variational approximation and
        the Gaussian process prior.

        For this variational family, we have
        ```math
        \begin{align}
        \operatorname{KL}[q(f(\cdot))\mid\mid p(\cdot)] & = \operatorname{KL}[q(u)\mid\mid p(u)]\\
            & = \operatorname{KL}[N(\mu  , S)\mid\mid N(0, I)].
        \end{align}
        ```

        Returns
        -------
            ScalarFloat: The KL-divergence between our variational
                approximation and the GP prior.
        """
        # Unpack variational parameters
        mu = self.variational_mean
        sqrt = cola.ops.Triangular(self.variational_root_covariance)

        # S = LLᵀ
        S = sqrt @ sqrt.T

        # Compute whitened KL divergence
        qu = GaussianDistribution(loc=jnp.atleast_1d(mu.squeeze()), scale=S)
        pu = GaussianDistribution(loc=jnp.zeros_like(jnp.atleast_1d(mu.squeeze())))
        return qu.kl_divergence(pu)

    def predict(self, test_inputs: Float[Array, "N D"]) -> GaussianDistribution:
        r"""Compute the predictive distribution of the GP at the test inputs t.

        This is the integral q(f(t)) = \int p(f(t)\midu) q(u) du, which can be computed in
        closed form as
        ```math
            \mathcal{N}\left(f(t); \mu t  +  \mathbf{K}_{tz} \mathbf{L}z^{\top} \mu  ,  \mathbf{K}_{tt}  -  \mathbf{K}_{tz} \mathbf{K}_{zz}^{-1} \mathbf{K}_{zt}  +  \mathbf{K}_{tz} \mathbf{L}z^{\top} S \mathbf{L}z^{-1} \mathbf{K}_{zt} \right).
        ```

        Args:
            test_inputs (Float[Array, "N D"]): The test inputs at which we wish to
                make a prediction.

        Returns
        -------
            GaussianDistribution: The predictive distribution of the low-rank GP at
                the test inputs.
        """
        # Unpack variational parameters
        mu = self.variational_mean
        sqrt = self.variational_root_covariance
        z = self.inducing_inputs

        # Unpack mean function and kernel
        mean_function = self.posterior.prior.mean_function
        kernel = self.posterior.prior.kernel

        Kzz = kernel.gram(z)
        Kzz += cola.ops.I_like(Kzz) * self.jitter
        Lz = lower_cholesky(Kzz)

        # Unpack test inputs
        t = test_inputs

        Ktt = kernel.gram(t)
        Kzt = kernel.cross_covariance(z, t)
        mut = mean_function(t)

        # Lz⁻¹ Kzt
        Lz_inv_Kzt = cola.solve(Lz, Kzt, Cholesky())

        # Ktz Lz⁻ᵀ sqrt
        Ktz_Lz_invT_sqrt = jnp.matmul(Lz_inv_Kzt.T, sqrt)

        # μt  +  Ktz Lz⁻ᵀ μ
        mean = mut + jnp.matmul(Lz_inv_Kzt.T, mu)

        # Ktt  -  Ktz Kzz⁻¹ Kzt  +  Ktz Lz⁻ᵀ S Lz⁻¹ Kzt  [recall S = sqrt sqrtᵀ]
        covariance = (
            Ktt
            - jnp.matmul(Lz_inv_Kzt.T, Lz_inv_Kzt)
            + jnp.matmul(Ktz_Lz_invT_sqrt, Ktz_Lz_invT_sqrt.T)
        )
        covariance += cola.ops.I_like(covariance) * self.jitter

        return GaussianDistribution(
            loc=jnp.atleast_1d(mean.squeeze()), scale=covariance
        )


@dataclass
class NaturalVariationalGaussian(AbstractVariationalGaussian):
    r"""The natural variational Gaussian family of probability distributions.

    The variational family is $`q(f(\cdot)) = \int p(f(\cdot)\mid u) q(u) \mathrm{d}u`$,
    where $`u = f(z)`$ are
    the function values at the inducing inputs $`z`$ and the distribution over the
    inducing inputs is $`q(u) = N(\mu, S)`$. Expressing the variational distribution, in
    the form of the exponential family, $`q(u) = exp(\theta^{\top} T(u) - a(\theta))`$, gives rise to the
    natural parameterisation $`\theta  = (\theta_{1}, \theta_{2}) = (S^{-1}\mu, -S^{-1}/2)`$, to perform model inference,
    where $`T(u) = [u, uu^{\top}]`$ are the sufficient statistics.
    """

    natural_vector: Float[Array, "M 1"] = None
    natural_matrix: Float[Array, "M M"] = None

    def __post_init__(self):
        if self.natural_vector is None:
            self.natural_vector = jnp.zeros((self.num_inducing, 1))

        if self.natural_matrix is None:
            self.natural_matrix = -0.5 * jnp.eye(self.num_inducing)

    def prior_kl(self) -> ScalarFloat:
        r"""Compute the KL-divergence between our current variational approximation
        and the Gaussian process prior.

        For this variational family, we have
        ```math
        \begin{align}
        \operatorname{KL}[q(f(\cdot))\mid\mid p(\cdot)] & = \operatorname{KL}[q(u)\mid\mid p(u)] \\
            & = \operatorname{KL}[N(\mu, S)\mid\mid N(mz, \mathbf{K}_{zz})],
        \end{align}
        ```
        with $\mu$ and $S$ computed from the natural parameterisation $\theta  = (S^{-1}\mu  , -S^{-1}/2)$.

        Returns
        -------
            ScalarFloat: The KL-divergence between our variational approximation and
                the GP prior.
        """
        # Unpack variational parameters
        natural_vector = self.natural_vector
        natural_matrix = self.natural_matrix
        z = self.inducing_inputs
        m = self.num_inducing

        # Unpack mean function and kernel
        mean_function = self.posterior.prior.mean_function
        kernel = self.posterior.prior.kernel

        # S⁻¹ = -2θ₂
        S_inv = -2 * natural_matrix
        S_inv += jnp.eye(m) * self.jitter

        # Compute L⁻¹, where LLᵀ = S, via a trick found in the NumPyro source code and https://nbviewer.org/gist/fehiepsi/5ef8e09e61604f10607380467eb82006#Precision-to-scale_tril:
        sqrt_inv = jnp.swapaxes(
            jnp.linalg.cholesky(S_inv[..., ::-1, ::-1])[..., ::-1, ::-1], -2, -1
        )

        # L = (L⁻¹)⁻¹I
        sqrt = jsp.linalg.solve_triangular(sqrt_inv, jnp.eye(m), lower=True)
        sqrt = cola.ops.Triangular(sqrt)

        # S = LLᵀ:
        S = sqrt @ sqrt.T

        # μ = Sθ₁
        mu = S @ natural_vector

        muz = mean_function(z)
        Kzz = kernel.gram(z)
        Kzz += cola.ops.I_like(Kzz) * self.jitter

        qu = GaussianDistribution(loc=jnp.atleast_1d(mu.squeeze()), scale=S)
        pu = GaussianDistribution(loc=jnp.atleast_1d(muz.squeeze()), scale=Kzz)

        return qu.kl_divergence(pu)

    def predict(self, test_inputs: Float[Array, "N D"]) -> GaussianDistribution:
        r"""Compute the predictive distribution of the GP at the test inputs $t$.

        This is the integral $`q(f(t)) = \int p(f(t)\mid u) q(u) \mathrm{d}u`$, which
        can be computed in closed form as
        ```math
             \mathcal{N}\left(f(t); \mu  t + \mathbf{K}_{tz} \mathbf{K}_{zz}^{-1} (\mu   - \mu  z),  \mathbf{K}_{tt} - \mathbf{K}_{tz} \mathbf{K}_{zz}^{-1} \mathbf{K}_{zt} + \mathbf{K}_{tz} \mathbf{K}_{zz}^{-1} S \mathbf{K}_{zz}^{-1} \mathbf{K}_{zt} \right),
        ```
        with $`\mu`$ and $`S`$ computed from the natural parameterisation
        $`\theta = (S^{-1}\mu  , -S^{-1}/2)`$.

        Returns
        -------
            GaussianDistribution: A function that accepts a set of test points and will
                return the predictive distribution at those points.
        """
        # Unpack variational parameters
        natural_vector = self.natural_vector
        natural_matrix = self.natural_matrix
        z = self.inducing_inputs
        m = self.num_inducing

        # Unpack mean function and kernel
        mean_function = self.posterior.prior.mean_function
        kernel = self.posterior.prior.kernel

        # S⁻¹ = -2θ₂
        S_inv = -2 * natural_matrix
        S_inv += jnp.eye(m) * self.jitter

        # Compute L⁻¹, where LLᵀ = S, via a trick found in the NumPyro source code and https://nbviewer.org/gist/fehiepsi/5ef8e09e61604f10607380467eb82006#Precision-to-scale_tril:
        sqrt_inv = jnp.swapaxes(
            jnp.linalg.cholesky(S_inv[..., ::-1, ::-1])[..., ::-1, ::-1], -2, -1
        )

        # L = (L⁻¹)⁻¹I
        sqrt = jsp.linalg.solve_triangular(sqrt_inv, jnp.eye(m), lower=True)

        # S = LLᵀ:
        S = jnp.matmul(sqrt, sqrt.T)

        # μ = Sθ₁
        mu = jnp.matmul(S, natural_vector)

        Kzz = kernel.gram(z)
        Kzz += cola.ops.I_like(Kzz) * self.jitter
        Lz = lower_cholesky(Kzz)
        muz = mean_function(z)

        Ktt = kernel.gram(test_inputs)
        Kzt = kernel.cross_covariance(z, test_inputs)
        mut = mean_function(test_inputs)

        # Lz⁻¹ Kzt
        Lz_inv_Kzt = cola.solve(Lz, Kzt, Cholesky())

        # Kzz⁻¹ Kzt
        Kzz_inv_Kzt = cola.solve(Lz.T, Lz_inv_Kzt, Cholesky())

        # Ktz Kzz⁻¹ L
        Ktz_Kzz_inv_L = jnp.matmul(Kzz_inv_Kzt.T, sqrt)

        # μt  +  Ktz Kzz⁻¹ (μ  -  μz)
        mean = mut + jnp.matmul(Kzz_inv_Kzt.T, mu - muz)

        # Ktt  -  Ktz Kzz⁻¹ Kzt  +  Ktz Kzz⁻¹ S Kzz⁻¹ Kzt  [recall S = LLᵀ]
        covariance = (
            Ktt
            - jnp.matmul(Lz_inv_Kzt.T, Lz_inv_Kzt)
            + jnp.matmul(Ktz_Kzz_inv_L, Ktz_Kzz_inv_L.T)
        )
        covariance += cola.ops.I_like(covariance) * self.jitter

        return GaussianDistribution(
            loc=jnp.atleast_1d(mean.squeeze()), scale=covariance
        )


@dataclass
class ExpectationVariationalGaussian(AbstractVariationalGaussian):
    r"""The natural variational Gaussian family of probability distributions.

    The variational family is $`q(f(\cdot)) = \int p(f(\cdot)\mid u) q(u) \mathrm{d}u`$, where $`u = f(z)`$ are the
    function values at the inducing inputs $`z`$ and the distribution over the inducing
    inputs is $`q(u) = \mathcal{N}(\mu, S)`$. Expressing the variational distribution, in the form of
    the exponential family, $`q(u) = exp(\theta^{\top} T(u) - a(\theta))`$, gives rise to the natural
    parameterisation $`\theta  = (\theta_{1}, \theta_{2}) = (S^{-1}\mu  , -S^{-1}/2)`$ and sufficient statistics
    $`T(u) = [u, uu^{\top}]`$. The expectation parameters are given by $`\nu = \int T(u) q(u) \mathrm{d}u`$.
    This gives a parameterisation, $`\nu = (\nu_{1}, \nu_{2}) = (\mu  , S + uu^{\top})`$ to perform model
    inference over.
    """

    expectation_vector: Float[Array, "M 1"] = None
    expectation_matrix: Float[Array, "M M"] = None

    def __post_init__(self):
        if self.expectation_vector is None:
            self.expectation_vector = jnp.zeros((self.num_inducing, 1))
        if self.expectation_matrix is None:
            self.expectation_matrix = jnp.eye(self.num_inducing)

    def prior_kl(self) -> ScalarFloat:
        r"""Evaluate the prior KL-divergence.

        Compute the KL-divergence between our current variational approximation and
        the Gaussian process prior.

        For this variational family, we have
        ```math
        \begin{align}
        \operatorname{KL}(q(f(\cdot))\mid\mid p(\cdot)) & = \operatorname{KL}(q(u)\mid\mid p(u)) \\
            & =\operatorname{KL}(\mathcal{N}(\mu, S)\mid\mid \mathcal{N}(m_z, K_{zz})),
        \end{align}
        ```
        where $\mu$ and $S$ are the expectation parameters of the variational
        distribution and $m_z$ and $K_{zz}$ are the mean and covariance of the prior
        distribution.

        Returns
        -------
            ScalarFloat: The KL-divergence between our variational approximation and
                the GP prior.
        """
        # Unpack variational parameters
        expectation_vector = self.expectation_vector
        expectation_matrix = self.expectation_matrix
        z = self.inducing_inputs

        # Unpack mean function and kernel
        mean_function = self.posterior.prior.mean_function
        kernel = self.posterior.prior.kernel

        # μ = η₁
        mu = expectation_vector

        # S = η₂ - η₁ η₁ᵀ
        S = expectation_matrix - jnp.outer(mu, mu)
        S = cola.ops.Dense(S)
        S = cola.PSD(S)
        S += cola.ops.I_like(S) * self.jitter

        muz = mean_function(z)
        Kzz = kernel.gram(z)
        Kzz += cola.ops.I_like(Kzz) * self.jitter

        qu = GaussianDistribution(loc=jnp.atleast_1d(mu.squeeze()), scale=S)
        pu = GaussianDistribution(loc=jnp.atleast_1d(muz.squeeze()), scale=Kzz)

        return qu.kl_divergence(pu)

    def predict(self, test_inputs: Float[Array, "N D"]) -> GaussianDistribution:
        r"""Evaluate the predictive distribution.

        Compute the predictive distribution of the GP at the test inputs $t$.

        This is the integral $q(f(t)) = \int p(f(t)\mid u)q(u)\mathrm{d}u$, which can
        be computed in closed form as  which can be computed in closed form as
        ```math
        \mathcal{N}(f(t); \mu_t + \mathbf{K}_{tz}\mathbf{K}_{zz}^{-1}(\mu - \mu_z), \mathbf{K}_{tt} - \mathbf{K}_{tz}\mathbf{K}_{zz}^{-1}\mathbf{K}_{zt} + \mathbf{K}_{tz}\mathbf{K}_{zz}^{-1}\mathbf{S} \mathbf{K}_{zz}^{-1}\mathbf{K}_{zt})
        ```

        with $\mu$ and $S$ computed from the expectation parameterisation
        $\eta = (\mu, S + uu^\top)$.

        Returns
        -------
            GaussianDistribution: The predictive distribution of the GP at the
                test inputs $t$.
        """
        # Unpack variational parameters
        expectation_vector = self.expectation_vector
        expectation_matrix = self.expectation_matrix
        z = self.inducing_inputs

        # Unpack mean function and kernel
        mean_function = self.posterior.prior.mean_function
        kernel = self.posterior.prior.kernel

        # μ = η₁
        mu = expectation_vector

        # S = η₂ - η₁ η₁ᵀ
        S = expectation_matrix - jnp.matmul(mu, mu.T)
        S = cola.ops.Dense(S)
        S += cola.ops.I_like(S) * self.jitter
        S = cola.PSD(S)

        # S = sqrt sqrtᵀ
        sqrt = lower_cholesky(S)

        Kzz = kernel.gram(z)
        Kzz += cola.ops.I_like(Kzz) * self.jitter
        Lz = lower_cholesky(Kzz)
        muz = mean_function(z)

        # Unpack test inputs
        t = test_inputs

        Ktt = kernel.gram(t)
        Kzt = kernel.cross_covariance(z, t)
        mut = mean_function(t)

        # Lz⁻¹ Kzt
        Lz_inv_Kzt = cola.solve(Lz, Kzt, Cholesky())

        # Kzz⁻¹ Kzt
        Kzz_inv_Kzt = cola.solve(Lz.T, Lz_inv_Kzt, Cholesky())

        # Ktz Kzz⁻¹ sqrt
        Ktz_Kzz_inv_sqrt = Kzz_inv_Kzt.T @ sqrt

        # μt  +  Ktz Kzz⁻¹ (μ  -  μz)
        mean = mut + jnp.matmul(Kzz_inv_Kzt.T, mu - muz)

        # Ktt  -  Ktz Kzz⁻¹ Kzt  +  Ktz Kzz⁻¹ S Kzz⁻¹ Kzt  [recall S = sqrt sqrtᵀ]
        covariance = (
            Ktt
            - jnp.matmul(Lz_inv_Kzt.T, Lz_inv_Kzt)
            + jnp.matmul(Ktz_Kzz_inv_sqrt, Ktz_Kzz_inv_sqrt.T)
        )
        covariance += cola.ops.I_like(covariance) * self.jitter

        return GaussianDistribution(
            loc=jnp.atleast_1d(mean.squeeze()), scale=covariance
        )


@dataclass
class CollapsedVariationalGaussian(AbstractVariationalGaussian):
    r"""Collapsed variational Gaussian.

    Collapsed variational Gaussian family of probability distributions.
    The key reference is Titsias, (2009) - Variational Learning of Inducing Variables
    in Sparse Gaussian Processes.
    """

    def __post_init__(self):
        if not isinstance(self.posterior.likelihood, Gaussian):
            raise TypeError("Likelihood must be Gaussian.")

    def predict(
        self, test_inputs: Float[Array, "N D"], train_data: Dataset
    ) -> GaussianDistribution:
        r"""Compute the predictive distribution of the GP at the test inputs.

        Args:
            test_inputs (Float[Array, "N D"]): The test inputs $t$ at which to make
                predictions.
            train_data (Dataset): The training data that was used to fit the GP.

        Returns
        -------
            GaussianDistribution: The predictive distribution of the collapsed
                variational Gaussian process at the test inputs $t$.
        """
        # Unpack test inputs
        t = test_inputs

        # Unpack training data
        x, y = train_data.X, train_data.y

        # Unpack variational parameters
        noise_var = self.posterior.likelihood.obs_stddev**2
        z = self.inducing_inputs
        m = self.num_inducing

        # Unpack mean function and kernel
        mean_function = self.posterior.prior.mean_function
        kernel = self.posterior.prior.kernel

        Kzx = kernel.cross_covariance(z, x)
        Kzz = kernel.gram(z)
        Kzz += cola.ops.I_like(Kzz) * self.jitter

        # Lz Lzᵀ = Kzz
        Lz = lower_cholesky(Kzz)

        # Lz⁻¹ Kzx
        Lz_inv_Kzx = cola.solve(Lz, Kzx, Cholesky())

        # A = Lz⁻¹ Kzt / o
        A = Lz_inv_Kzx / self.posterior.likelihood.obs_stddev

        # AAᵀ
        AAT = jnp.matmul(A, A.T)

        # LLᵀ = I + AAᵀ
        L = jnp.linalg.cholesky(jnp.eye(m) + AAT)

        mux = mean_function(x)
        diff = y - mux

        # Lz⁻¹ Kzx (y - μx)
        Lz_inv_Kzx_diff = jsp.linalg.cho_solve((L, True), jnp.matmul(Lz_inv_Kzx, diff))

        # Kzz⁻¹ Kzx (y - μx)
        Kzz_inv_Kzx_diff = cola.solve(Lz.T, Lz_inv_Kzx_diff, Cholesky())

        Ktt = kernel.gram(t)
        Kzt = kernel.cross_covariance(z, t)
        mut = mean_function(t)

        # Lz⁻¹ Kzt
        Lz_inv_Kzt = cola.solve(Lz, Kzt, Cholesky())

        # L⁻¹ Lz⁻¹ Kzt
        L_inv_Lz_inv_Kzt = jsp.linalg.solve_triangular(L, Lz_inv_Kzt, lower=True)

        # μt + 1/o² Ktz Kzz⁻¹ Kzx (y - μx)
        mean = mut + jnp.matmul(Kzt.T / noise_var, Kzz_inv_Kzx_diff)

        # Ktt  -  Ktz Kzz⁻¹ Kzt  +  Ktz Lz⁻¹ (I + AAᵀ)⁻¹ Lz⁻¹ Kzt
        covariance = (
            Ktt
            - jnp.matmul(Lz_inv_Kzt.T, Lz_inv_Kzt)
            + jnp.matmul(L_inv_Lz_inv_Kzt.T, L_inv_Lz_inv_Kzt)
        )
        covariance += cola.ops.I_like(covariance) * self.jitter

        return GaussianDistribution(
            loc=jnp.atleast_1d(mean.squeeze()), scale=covariance
        )


__all__ = [
    "AbstractVariationalFamily",
    "AbstractVariationalGaussian",
    "VariationalGaussian",
    "WhitenedVariationalGaussian",
    "NaturalVariationalGaussian",
    "ExpectationVariationalGaussian",
    "CollapsedVariationalGaussian",
]



# Copyright 2023 The JaxGaussianProcesses Contributors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from dataclasses import dataclass

from beartype.typing import (
    Callable,
    Optional,
)
import optax as ox

import gpjax as gpx
from gpjax.dataset import Dataset
from gpjax.gps import (
    AbstractLikelihood,
    AbstractPosterior,
    AbstractPrior,
)
from gpjax.objectives import AbstractObjective
from gpjax.typing import KeyArray

LikelihoodBuilder = Callable[[int], AbstractLikelihood]
"""Type alias for likelihood builders, which take the number of datapoints as input and
return a likelihood object initialised with the given number of datapoints."""


@dataclass
class PosteriorHandler:
    """
    Class for handling the creation and updating of a GP posterior as new data is
    observed.

    Attributes:
        prior (AbstractPrior): Prior to use when forming the posterior.
        likelihood_builder (LikelihoodBuilder): Function which takes the number of
        datapoints as input and returns a likelihood object initialised with the given
        number of datapoints.
        optimization_objective (AbstractObjective): Objective to use for optimizing the
        posterior hyperparameters.
        optimizer (ox.GradientTransformation): Optax optimizer to use for optimizing the
        posterior hyperparameters.
        num_optimization_iterations (int): Number of iterations to optimize
        the posterior hyperparameters for.
    """

    prior: AbstractPrior
    likelihood_builder: LikelihoodBuilder
    optimization_objective: AbstractObjective
    optimizer: ox.GradientTransformation
    num_optimization_iters: int

    def __post_init__(self):
        if self.num_optimization_iters < 1:
            raise ValueError("num_optimization_iters must be greater than 0.")

    def get_posterior(
        self, dataset: Dataset, optimize: bool, key: Optional[KeyArray] = None
    ) -> AbstractPosterior:
        """
        Initialise (and optionally optimize) a posterior using the given dataset.

        Args:
            dataset (Dataset): Dataset to get posterior for.
            optimize (bool): Whether to optimize the posterior hyperparameters.
            key (Optional[KeyArray]): A JAX PRNG key which is used for optimizing the posterior
            hyperparameters.

        Returns:
            Posterior for the given dataset.
        """
        posterior = self.prior * self.likelihood_builder(dataset.n)

        if optimize:
            if key is None:
                raise ValueError(
                    "A key must be provided in order to optimize the posterior."
                )
            posterior = self._optimize_posterior(posterior, dataset, key)

        return posterior

    def update_posterior(
        self,
        dataset: Dataset,
        previous_posterior: AbstractPosterior,
        optimize: bool,
        key: Optional[KeyArray] = None,
    ) -> AbstractPosterior:
        """
        Update the given posterior with the given dataset. This needs to be done when
        the number of datapoints in the (training) dataset of the posterior changes, as
        the `AbstractLikelihood` class requires the number of datapoints to be specified.
        Hyperparameters may or may not be optimized, depending on the value of the
        `optimize` parameter. Note that the updated poterior will be initialised with
        the same prior hyperparameters as the previous posterior, but the likelihood
        will be re-initialised with the new number of datapoints, and hyperparameters
        set as in the `likelihood_builder` function.

        Args:
            dataset: Dataset to get posterior for.
            previous_posterior: Posterior being updated. This is supplied as one may
            wish to simply increase the number of datapoints in the likelihood, without
            optimizing the posterior hyperparameters, in which case the previous
            posterior can be used to obtain the previously set prior hyperparameters.
            optimize: Whether to optimize the posterior hyperparameters.
            key: A JAX PRNG key which is used for optimizing the posterior
            hyperparameters.
        """
        posterior = previous_posterior.prior * self.likelihood_builder(dataset.n)

        if optimize:
            if key is None:
                raise ValueError(
                    "A key must be provided in order to optimize the posterior."
                )
            posterior = self._optimize_posterior(posterior, dataset, key)
        return posterior

    def _optimize_posterior(
        self, posterior: AbstractPosterior, dataset: Dataset, key: KeyArray
    ) -> AbstractPosterior:
        """
        Takes a posterior and corresponding dataset and optimizes the posterior using the
        GPJax `fit` method.

        Args:
            posterior: Posterior being optimized.
            dataset: Dataset used for optimizing posterior.
            key: A JAX PRNG key for generating random numbers.
        Returns:
            Optimized posterior.
        """
        opt_posterior, _ = gpx.fit(
            model=posterior,
            objective=self.optimization_objective,
            train_data=dataset,
            optim=self.optimizer,
            num_iters=self.num_optimization_iters,
            safe=True,
            key=key,
            verbose=False,
        )

        return opt_posterior



# Copyright 2022 The JaxGaussianProcesses Contributors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================


from dataclasses import dataclass
from typing import (
    NamedTuple,
    Union,
)

import jax.numpy as jnp
from jaxtyping import (
    Float,
    Int,
)
import tensorflow_probability.substrates.jax as tfp

from gpjax.base import (
    param_field,
    static_field,
)
from gpjax.kernels.base import AbstractKernel
from gpjax.typing import (
    Array,
    ScalarInt,
)

tfb = tfp.bijectors

CatKernelParams = NamedTuple(
    "CatKernelParams",
    [("stddev", Float[Array, "N 1"]), ("cholesky_lower", Float[Array, " N*(N-1)//2"])],
)


@dataclass
class CatKernel(AbstractKernel):
    r"""The categorical kernel is defined for a fixed number of values of categorical input.

    It stores a standard dev for each input value (i.e. the diagonal of the gram), and a lower cholesky factor for correlations.
    It returns the corresponding values from an the gram matrix when called.

    Args:
        stddev (Float[Array, "N"]): The standard deviation parameters, one for each input space value.
        cholesky_lower (Float[Array, "N*(N-1)//2 N"]): The parameters for the Cholesky factor of the gram matrix.
        inspace_vals (list): The values in the input space this CatKernel works for. Stored for order reference, making clear the indices used for each input space value.
        name (str): The name of the kernel.
        input_1hot (bool): If True, the kernel expect to be called with a 1-hot encoding of the input space values. If False, it expects the indices of the input space values.

    Raises:
        ValueError: If the number of diagonal variance parameters does not match the number of input space values.
    """

    stddev: Float[Array, " N"] = param_field(jnp.ones((2,)), bijector=tfb.Softplus())
    cholesky_lower: Float[Array, "N N"] = param_field(
        jnp.eye(2), bijector=tfb.CorrelationCholesky()
    )
    inspace_vals: Union[list, None] = static_field(None)
    name: str = "Categorical Kernel"
    input_1hot: bool = static_field(False)

    def __post_init__(self):
        if self.inspace_vals is not None and len(self.inspace_vals) != len(self.stddev):
            raise ValueError(
                f"The number of stddev parameters ({len(self.stddev)}) has to match the number of input space values ({len(self.inspace_vals)}), unless inspace_vals is None."
            )

    @property
    def explicit_gram(self) -> Float[Array, "N N"]:
        """Access the PSD gram matrix resulting from the parameters.

        Returns:
            Float[Array, "N N"]: The gram matrix.
        """
        L = self.stddev.reshape(-1, 1) * self.cholesky_lower
        return L @ L.T

    def __call__(  # TODO not consistent with general kernel interface
        self,
        x: Union[ScalarInt, Int[Array, " N"]],
        y: Union[ScalarInt, Int[Array, " N"]],
    ):
        r"""Compute the (co)variance between a pair of dictionary indices.

        Args:
            x (Union[ScalarInt, Int[Array, "N"]]): The index of the first dictionary entry, or its one-hot encoding.
            y (Union[ScalarInt, Int[Array, "N"]]): The index of the second dictionary entry, or its one-hot encoding.

        Returns
        -------
            ScalarFloat: The value of $k(v_i, v_j)$.
        """
        try:
            x = x.squeeze()
            y = y.squeeze()
        except AttributeError:
            pass
        if self.input_1hot:
            return self.explicit_gram[jnp.outer(x, y) == 1]
        else:
            return self.explicit_gram[x, y]

    @staticmethod
    def num_cholesky_lower_params(num_inspace_vals: ScalarInt) -> ScalarInt:
        """Compute the number of parameters required to store the lower triangular Cholesky factor of the gram matrix.

        Args:
            num_inspace_vals (ScalarInt): The number of values in the input space.

        Returns:
            ScalarInt: The number of parameters required to store the lower triangle of the Cholesky factor of the gram matrix.
        """
        return num_inspace_vals * (num_inspace_vals - 1) // 2

    @staticmethod
    def gram_to_stddev_cholesky_lower(gram: Float[Array, "N N"]) -> CatKernelParams:
        """Compute the standard deviation and lower triangular Cholesky factor of the gram matrix.

        Args:
            gram (Float[Array, "N N"]): The gram matrix.

        Returns:
            tuple[Float[Array, "N"], Float[Array, "N N"]]: The standard deviation and lower triangular Cholesky factor of the gram matrix, where the latter is scaled to result in unit variances.
        """
        stddev = jnp.sqrt(jnp.diag(gram))
        L = jnp.linalg.cholesky(gram) / stddev.reshape(-1, 1)
        return CatKernelParams(stddev, L)



# Copyright 2023 The JaxGaussianProcesses Contributors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from abc import (
    ABC,
    abstractmethod,
)
from dataclasses import dataclass

from beartype.typing import (
    Callable,
    Mapping,
)

from gpjax.dataset import Dataset
from gpjax.decision_making.utils import OBJECTIVE
from gpjax.gps import AbstractPosterior
from gpjax.typing import (
    Array,
    Float,
    KeyArray,
)

SinglePointUtilityFunction = Callable[[Float[Array, "N D"]], Float[Array, "N 1"]]
"""
Type alias for utility functions which don't support batching, and instead characterise
the utility of querying a single point, rather than a batch of points. They take an array of points of shape $`[N, D]`$
and return the value of the utility function at each point in an array of shape $`[N, 1]`$.
"""


UtilityFunction = SinglePointUtilityFunction
"""
Type alias for all utility functions. Currently we only support
`SinglePointUtilityFunction`s, but in future may support batched utility functions too.
Note that `UtilityFunction`s are *maximised* in order to decide which point, or batch of points, to query next.
"""


@dataclass
class AbstractSinglePointUtilityFunctionBuilder(ABC):
    """
    Abstract class for building utility functions which don't support batches. As such,
    they characterise the utility of querying a single point next.
    """

    def check_objective_present(
        self,
        posteriors: Mapping[str, AbstractPosterior],
        datasets: Mapping[str, Dataset],
    ) -> None:
        """
        Check that the objective posterior and dataset are present in the posteriors and
        datasets.

        Args:
            posteriors (Mapping[str, AbstractPosterior]): Dictionary of posteriors to be
            used to form the utility function.
            datasets (Mapping[str, Dataset]): Dictionary of datasets which may be used
            to form the utility function.

        Raises:
            ValueError: If the objective posterior or dataset are not present in the
            posteriors or datasets.
        """
        if OBJECTIVE not in posteriors.keys():
            raise ValueError("Objective posterior not found in posteriors")
        elif OBJECTIVE not in datasets.keys():
            raise ValueError("Objective dataset not found in datasets")

    @abstractmethod
    def build_utility_function(
        self,
        posteriors: Mapping[str, AbstractPosterior],
        datasets: Mapping[str, Dataset],
        key: KeyArray,
    ) -> SinglePointUtilityFunction:
        """
        Build a `UtilityFunction` from a set of posteriors and datasets.

        Args:
            posteriors (Mapping[str, AbstractPosterior]): Dictionary of posteriors to be
            used to form the utility function.
            datasets (Mapping[str, Dataset]): Dictionary of datasets which may be used
            to form the utility function.
            key (KeyArray): JAX PRNG key used for random number generation.

        Returns:
            SinglePointUtilityFunction: Utility function to be *maximised* in order to
            decide which point to query next.
        """
        raise NotImplementedError


AbstractUtilityFunctionBuilder = AbstractSinglePointUtilityFunctionBuilder
"""
Type alias for utility function builders. For now this only include single point utility
function builders, but in the future we may support batched utility function builders.
"""



# Copyright 2023 The GPJax Contributors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from abc import (
    ABC,
    abstractmethod,
)
from dataclasses import dataclass

import jax.numpy as jnp
from jaxtyping import (
    Array,
    Float,
)

from gpjax.dataset import Dataset
from gpjax.decision_making.search_space import ContinuousSearchSpace
from gpjax.typing import KeyArray


class AbstractContinuousTestFunction(ABC):
    """
    Abstract base class for continuous test functions.

    Attributes:
        search_space (ContinuousSearchSpace): Search space for the function.
        minimizer (Float[Array, '1 D']): Minimizer of the function (to 5 decimal places)
        minimum (Float[Array, '1 1']): Minimum of the function (to 5 decimal places).
    """

    search_space: ContinuousSearchSpace
    minimizer: Float[Array, "1 D"]
    minimum: Float[Array, "1 1"]

    def generate_dataset(self, num_points: int, key: KeyArray) -> Dataset:
        """
        Generate a toy dataset from the test function.

        Args:
            num_points (int): Number of points to sample.
            key (KeyArray): JAX PRNG key.

        Returns:
            Dataset: Dataset of points sampled from the test function.
        """
        X = self.search_space.sample(num_points=num_points, key=key)
        y = self.evaluate(X)
        return Dataset(X=X, y=y)

    def generate_test_points(
        self, num_points: int, key: KeyArray
    ) -> Float[Array, "N D"]:
        """
        Generate test points from the search space of the test function.

        Args:
            num_points (int): Number of points to sample.
            key (KeyArray): JAX PRNG key.

        Returns:
            Float[Array, 'N D']: Test points sampled from the search space.
        """
        return self.search_space.sample(num_points=num_points, key=key)

    @abstractmethod
    def evaluate(self, x: Float[Array, "N D"]) -> Float[Array, "N 1"]:
        """
        Evaluate the test function at a set of points.

        Args:
            x (Float[Array, 'N D']): Points to evaluate the test function at.

        Returns:
            Float[Array, 'N 1']: Values of the test function at the points.
        """
        raise NotImplementedError


@dataclass
class Forrester(AbstractContinuousTestFunction):
    """
    Forrester function introduced in 'Engineering design via surrogate modelling: a
    practical guide' (Forrester et al. 2008), rescaled to have zero mean and unit
    variance over $`[0, 1]`$.
    """

    search_space = ContinuousSearchSpace(
        lower_bounds=jnp.array([0.0]),
        upper_bounds=jnp.array([1.0]),
    )
    minimizer = jnp.array([[0.75725]])
    minimum = jnp.array([[-1.45280]])

    def evaluate(self, x: Float[Array, "N D"]) -> Float[Array, "N 1"]:
        mean = 0.45321
        std = jnp.sqrt(19.8577)
        return (((6 * x - 2) ** 2) * jnp.sin(12 * x - 4) - mean) / std


@dataclass
class LogarithmicGoldsteinPrice(AbstractContinuousTestFunction):
    """
    Logarithmic Goldstein-Price function introduced in 'A benchmark of kriging-based
    infill criteria for noisy optimization' (Picheny et al. 2013), which has zero mean
    and unit variance over $`[0, 1]^2`$.
    """

    search_space = ContinuousSearchSpace(
        lower_bounds=jnp.array([0.0, 0.0]),
        upper_bounds=jnp.array([1.0, 1.0]),
    )
    minimizer = jnp.array([[0.5, 0.25]])
    minimum = jnp.array([[-3.12913]])

    def evaluate(self, x: Float[Array, "N D"]) -> Float[Array, "N 1"]:
        x1 = 4.0 * x[:, 0] - 2.0
        x2 = 4.0 * x[:, 1] - 2.0
        a = 1.0 + (x1 + x2 + 1.0) ** 2 * (
            19.0
            - 14.0 * x1
            + 3.0 * (x1**2)
            - 14.0 * x2
            + 6.0 * x1 * x2
            + 3.0 * (x2**2)
        )
        b = 30.0 + (2.0 * x1 - 3.0 * x2) ** 2 * (
            18.0
            - 32.0 * x1
            + 12.0 * (x1**2)
            + 48.0 * x2
            - 36.0 * x1 * x2
            + 27.0 * (x2**2)
        )
        return ((jnp.log((a * b)) - 8.693) / 2.427).reshape(-1, 1)


@dataclass
class Quadratic(AbstractContinuousTestFunction):
    """
    Toy quadratic function defined over $`[0, 1]`$.
    """

    search_space = ContinuousSearchSpace(
        lower_bounds=jnp.array([0.0]),
        upper_bounds=jnp.array([1.0]),
    )
    minimizer = jnp.array([[0.5]])
    minimum = jnp.array([[0.0]])

    def evaluate(self, x: Float[Array, "N D"]) -> Float[Array, "N 1"]:
        return (x - 0.5) ** 2



# Copyright 2022 The JaxGaussianProcesses Contributors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import abc
from dataclasses import dataclass
import typing as tp

from cola import PSD
from cola.ops import (
    Dense,
    Diagonal,
    LinearOperator,
)
from jax import vmap
from jaxtyping import (
    Float,
    Num,
)

from gpjax.typing import Array

Kernel = tp.TypeVar("Kernel", bound="gpjax.kernels.base.AbstractKernel")  # noqa: F821


@dataclass
class AbstractKernelComputation:
    r"""Abstract class for kernel computations."""

    def gram(
        self,
        kernel: Kernel,
        x: Num[Array, "N D"],
    ) -> LinearOperator:
        r"""Compute Gram covariance operator of the kernel function.

        Args:
            kernel (AbstractKernel): the kernel function.
            x (Float[Array, "N N"]): The inputs to the kernel function.

        Returns
        -------
            LinearOperator: Gram covariance operator of the kernel function.
        """
        Kxx = self.cross_covariance(kernel, x, x)
        return PSD(Dense(Kxx))

    @abc.abstractmethod
    def cross_covariance(
        self, kernel: Kernel, x: Num[Array, "N D"], y: Num[Array, "M D"]
    ) -> Float[Array, "N M"]:
        r"""For a given kernel, compute the NxM gram matrix on an a pair
        of input matrices with shape NxD and MxD.

        Args:
            kernel (AbstractKernel): the kernel function.
            x (Float[Array,"N D"]): The first input matrix.
            y (Float[Array,"M D"]): The second input matrix.

        Returns
        -------
            Float[Array, "N M"]: The computed cross-covariance.
        """
        raise NotImplementedError

    def diagonal(self, kernel: Kernel, inputs: Num[Array, "N D"]) -> Diagonal:
        r"""For a given kernel, compute the elementwise diagonal of the
        NxN gram matrix on an input matrix of shape NxD.

        Args:
            kernel (AbstractKernel): the kernel function.
            inputs (Float[Array, "N D"]): The input matrix.

        Returns
        -------
            Diagonal: The computed diagonal variance entries.
        """
        return PSD(Diagonal(diag=vmap(lambda x: kernel(x, x))(inputs)))



# Copyright 2023 The GPJax Contributors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import cola
import jax.numpy as jnp

# TODO: Once this functionality is supported in CoLA, remove this.


@cola.dispatch
def lower_cholesky(A: cola.ops.LinearOperator):  # noqa: F811
    """Returns the lower Cholesky factor of a linear operator.

    Args:
        A (cola.ops.LinearOperator): A linear operator.

    Returns:
        cola.ops.LinearOperator: The lower Cholesky factor of A.
    """

    if cola.PSD not in A.annotations:
        raise ValueError(
            "Expected LinearOperator to be PSD, did you forget to use cola.PSD?"
        )

    return cola.ops.Triangular(jnp.linalg.cholesky(A.to_dense()), lower=True)


@lower_cholesky.dispatch
def _(A: cola.ops.Diagonal):  # noqa: F811
    return cola.ops.Diagonal(jnp.sqrt(A.diag))


@lower_cholesky.dispatch
def _(A: cola.ops.Identity):  # noqa: F811
    return A


@lower_cholesky.dispatch
def _(A: cola.ops.Kronecker):  # noqa: F811
    return cola.ops.Kronecker(*[lower_cholesky(Ai) for Ai in A.Ms])


@lower_cholesky.dispatch
def _(A: cola.ops.BlockDiag):  # noqa: F811
    return cola.ops.BlockDiag(
        *[lower_cholesky(Ai) for Ai in A.Ms], multiplicities=A.multiplicities
    )



# Copyright 2023 The JaxGaussianProcesses Contributors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from abc import (
    ABC,
    abstractmethod,
)
from dataclasses import dataclass

import jax.numpy as jnp
import jax.random as jr
from jaxopt import ScipyBoundedMinimize

from gpjax.decision_making.search_space import (
    AbstractSearchSpace,
    ContinuousSearchSpace,
)
from gpjax.decision_making.utility_functions import SinglePointUtilityFunction
from gpjax.typing import (
    Array,
    Float,
    KeyArray,
    ScalarFloat,
)


def _get_discrete_maximizer(
    query_points: Float[Array, "N D"], utility_function: SinglePointUtilityFunction
) -> Float[Array, "1 D"]:
    """Get the point which maximises the utility function evaluated at a given set of points.

    Args:
        query_points (Float[Array, "N D"]): Set of points at which to evaluate the
        utility function.
        utility_function (SinglePointUtilityFunction): Single point utility function to
        be evaluated at "query_points".

    Returns:
        Float[Array, "1 D"]: Point in `query_points` which maximises the utility function.
    """
    utility_function_values = utility_function(query_points)
    max_utility_function_value_idx = jnp.argmax(
        utility_function_values, axis=0, keepdims=True
    )
    best_sample_point = jnp.take_along_axis(
        query_points, max_utility_function_value_idx, axis=0
    )
    return best_sample_point


@dataclass
class AbstractSinglePointUtilityMaximizer(ABC):
    """Abstract base class for single point utility function maximizers."""

    @abstractmethod
    def maximize(
        self,
        utility_function: SinglePointUtilityFunction,
        search_space: AbstractSearchSpace,
        key: KeyArray,
    ) -> Float[Array, "1 D"]:
        """Maximize the given utility function over the search space provided.

        Args:
            utility_function (UtilityFunction): Utility function to be
            maximized.
            search_space (AbstractSearchSpace): Search space over which to maximize
            the utility function.
            key (KeyArray): JAX PRNG key.

        Returns:
            Float[Array, "1 D"]: Point at which the utility function is maximized.
        """
        raise NotImplementedError


@dataclass
class ContinuousSinglePointUtilityMaximizer(AbstractSinglePointUtilityMaximizer):
    """The `ContinuousUtilityMaximizer` class is used to maximize utility
    functions over the continuous domain with L-BFGS-B. First we sample the utility
    function at `num_initial_samples` points from the search space, and then we run
    L-BFGS-B from the best of these initial points. We run this process `num_restarts`
    number of times, each time sampling a different random set of
    `num_initial_samples`initial points.
    """

    num_initial_samples: int
    num_restarts: int

    def __post_init__(self):
        if self.num_initial_samples < 1:
            raise ValueError(
                f"num_initial_samples must be greater than 0, got {self.num_initial_samples}."
            )
        elif self.num_restarts < 1:
            raise ValueError(
                f"num_restarts must be greater than 0, got {self.num_restarts}."
            )

    def maximize(
        self,
        utility_function: SinglePointUtilityFunction,
        search_space: ContinuousSearchSpace,
        key: KeyArray,
    ) -> Float[Array, "1 D"]:
        max_observed_utility_function_value = None
        maximizer = None

        for _ in range(self.num_restarts):
            key, _ = jr.split(key)
            initial_sample_points = search_space.sample(
                self.num_initial_samples, key=key
            )
            best_initial_sample_point = _get_discrete_maximizer(
                initial_sample_points, utility_function
            )

            def _scalar_utility_function(x: Float[Array, "1 D"]) -> ScalarFloat:
                """
                The Jaxopt minimizer requires a function which returns a scalar. It calls the
                utility function with one point at a time, so the utility function
                returns an array of shape [1, 1], so  we index to return a scalar. Note that
                we also return the negative of the utility function - this is because
                utility functions should be *maximimized* but the Jaxopt minimizer
                minimizes functions.
                """
                return -utility_function(x)[0][0]

            lbfgsb = ScipyBoundedMinimize(
                fun=_scalar_utility_function, method="l-bfgs-b"
            )
            bounds = (search_space.lower_bounds, search_space.upper_bounds)
            optimized_point = lbfgsb.run(
                best_initial_sample_point, bounds=bounds
            ).params
            optimized_utility_function_value = _scalar_utility_function(optimized_point)
            if (max_observed_utility_function_value is None) or (
                optimized_utility_function_value > max_observed_utility_function_value
            ):
                max_observed_utility_function_value = optimized_utility_function_value
                maximizer = optimized_point
        return maximizer


AbstractUtilityMaximizer = AbstractSinglePointUtilityMaximizer
"""
Type alias for a utility maximizer. Currently we only support single point utility
functions, but in future may support batched utility functions.
"""



# Copyright 2022 The JaxGaussianProcesses Contributors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

from jaxtyping import (
    Int,
    Num,
)

from gpjax.typing import Array


def jax_gather_nd(
    params: Num[Array, " N *rest"], indices: Int[Array, " M 1"]
) -> Num[Array, " M *rest"]:
    r"""Slice a `params` array at a set of `indices`.

    This is a reimplementation of TensorFlow's `gather_nd` function:
    [link](https://www.tensorflow.org/api_docs/python/tf/gather_nd)

    Args:
        params (Num[Array]): An arbitrary array with leading axes of length $N$ upon
            which we shall slice.
        indices (Float[Int]): An integer array of length $M$ with values in the range
            $[0, N)$ whose value at index $i$ will be used to slice `params` at
            index $i$.

    Returns
    -------
        Num[Array: An arbitrary array with leading axes of length $M$.
    """
    tuple_indices = tuple(indices[..., i] for i in range(indices.shape[-1]))
    return params[tuple_indices]



# Copyright 2023 The JaxGaussianProcesses Contributors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from abc import (
    ABC,
    abstractmethod,
)
from dataclasses import dataclass

from jaxtyping import Float
import tensorflow_probability.substrates.jax as tfp

from gpjax.typing import (
    Array,
    KeyArray,
)


@dataclass
class AbstractSearchSpace(ABC):
    """The `AbstractSearchSpace` class is an abstract base class for
    search spaces, which are used to define domains for sampling and optimisation functionality in GPJax.
    """

    @abstractmethod
    def sample(self, num_points: int, key: KeyArray) -> Float[Array, "N D"]:
        """Sample points from the search space.
        Args:
            num_points (int): Number of points to be sampled from the search space.
            key (KeyArray): JAX PRNG key.
        Returns:
            Float[Array, "N D"]: `num_points` points sampled from the search space.
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def dimensionality(self) -> int:
        """Dimensionality of the search space.
        Returns:
            int: Dimensionality of the search space.
        """
        raise NotImplementedError


@dataclass
class ContinuousSearchSpace(AbstractSearchSpace):
    """The `ContinuousSearchSpace` class is used to bound the domain of continuous real functions of dimension $`D`$."""

    lower_bounds: Float[Array, " D"]
    upper_bounds: Float[Array, " D"]

    def __post_init__(self):
        if not self.lower_bounds.dtype == self.upper_bounds.dtype:
            raise ValueError("Lower and upper bounds must have the same dtype.")
        if self.lower_bounds.shape != self.upper_bounds.shape:
            raise ValueError("Lower and upper bounds must have the same shape.")
        if self.lower_bounds.shape[0] == 0:
            raise ValueError("Lower and upper bounds cannot be empty")
        if not (self.lower_bounds <= self.upper_bounds).all():
            raise ValueError("Lower bounds must be less than upper bounds.")

    @property
    def dimensionality(self) -> int:
        return self.lower_bounds.shape[0]

    def sample(self, num_points: int, key: KeyArray) -> Float[Array, "N D"]:
        """Sample points from the search space using a Halton sequence.

        Args:
            num_points (int): Number of points to be sampled from the search space.
            key (KeyArray): JAX PRNG key.
        Returns:
            Float[Array, "N D"]: `num_points` points sampled using the Halton sequence
            from the search space.
        """
        if num_points <= 0:
            raise ValueError("Number of points must be greater than 0.")

        initial_sample = tfp.mcmc.sample_halton_sequence(
            dim=self.dimensionality, num_results=num_points, seed=key
        )
        return (
            self.lower_bounds + (self.upper_bounds - self.lower_bounds) * initial_sample
        )



# Copyright 2022 The GPJax Contributors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

from abc import abstractmethod
from dataclasses import (
    dataclass,
    field,
)
from typing import overload

from beartype.typing import (
    Any,
    Callable,
    Optional,
    Union,
)
import cola
from cola.linalg.decompositions.decompositions import Cholesky
from cola.ops import Dense
import jax.numpy as jnp
from jax.random import (
    PRNGKey,
    normal,
)
from jaxtyping import (
    Float,
    Num,
)

from gpjax.base import (
    Module,
    param_field,
    static_field,
)
from gpjax.dataset import Dataset
from gpjax.distributions import (
    GaussianDistribution,
    ReshapedDistribution,
    ReshapedGaussianDistribution,
)
from gpjax.kernels import (
    RFF,
    White,
)
from gpjax.kernels.base import AbstractKernel
from gpjax.likelihoods import (
    AbstractLikelihood,
    Gaussian,
    NonGaussianLikelihood,
)
from gpjax.lower_cholesky import lower_cholesky
from gpjax.mean_functions import AbstractMeanFunction
from gpjax.typing import (
    Array,
    FunctionalSample,
    KeyArray,
)


@dataclass
class AbstractPrior(Module):
    r"""Abstract Gaussian process prior."""

    kernel: AbstractKernel
    mean_function: AbstractMeanFunction
    jitter: float = static_field(1e-6)

    # TODO: when letting kernels be responsible for certain features, like
    # RBF(features=["outp_idx"]), this can be folded into the kernel,
    # just not sure how to ensure Kronecker structure then
    out_kernel: AbstractKernel = field(default_factory=White)

    def __call__(self, *args: Any, **kwargs: Any) -> ReshapedGaussianDistribution:
        r"""Evaluate the Gaussian process at the given points.

        The output of this function is a
        [TensorFlow probability distribution](https://www.tensorflow.org/probability/api_docs/python/tfp/substrates/jax/distributions) from which the
        the latent function's mean and covariance can be evaluated and the distribution
        can be sampled.

        Under the hood, `__call__` is calling the objects `predict` method. For this
        reasons, classes inheriting the `AbstractPrior` class, should not overwrite the
        `__call__` method and should instead define a `predict` method.

        Args:
            *args (Any): The arguments to pass to the GP's `predict` method.
            **kwargs (Any): The keyword arguments to pass to the GP's `predict` method.

        Returns
        -------
            ReshapedGaussianDistribution: A multivariate normal random variable representation
                of the Gaussian process, possibly with reshaped events.
        """
        return self.predict(*args, **kwargs)

    @abstractmethod
    def predict(self, *args: Any, **kwargs: Any) -> ReshapedGaussianDistribution:
        r"""Evaluate the predictive distribution.

        Compute the latent function's multivariate normal distribution for a
        given set of parameters. For any class inheriting the `AbstractPrior` class,
        this method must be implemented.

        Args:
            *args (Any): Arguments to the predict method.
            **kwargs (Any): Keyword arguments to the predict method.

        Returns
        -------
            ReshapedGaussianDistribution: A multivariate normal random variable representation
                of the Gaussian process, possibly with reshaped events.
        """
        raise NotImplementedError


#######################
# GP Priors
#######################
@dataclass
class Prior(AbstractPrior):
    r"""A Gaussian process prior object.

    The GP is parameterised by a
    [mean](https://docs.jaxgaussianprocesses.com/api/mean_functions/)
    and [kernel](https://docs.jaxgaussianprocesses.com/api/kernels/base/)
    function.

    A Gaussian process prior parameterised by a mean function $`m(\cdot)`$ and a kernel
    function $`k(\cdot, \cdot)`$ is given by
    $`p(f(\cdot)) = \mathcal{GP}(m(\cdot), k(\cdot, \cdot))`$.

    To invoke a `Prior` distribution, a kernel and mean function must be specified.

    Example:
    ```python
        >>> import gpjax as gpx

        >>> kernel = gpx.kernels.RBF()
        >>> meanf = gpx.mean_functions.Zero()
        >>> prior = gpx.Prior(mean_function=meanf, kernel = kernel)
    ```
    """

    # @overload
    # def __mul__(self, other: Gaussian) -> "ConjugatePosterior":
    #     ...

    # @overload
    # def __mul__(self, other: NonGaussianLikelihood) -> "NonConjugatePosterior":
    #     ...

    # @overload
    # def __mul__(self, other: AbstractLikelihood) -> "AbstractPosterior":
    #     ...

    def __mul__(self, other):
        r"""Combine the prior with a likelihood to form a posterior distribution.

        The product of a prior and likelihood is proportional to the posterior
        distribution. By computing the product of a GP prior and a likelihood
        object, a posterior GP object will be returned. Mathematically, this can
        be described by:
        ```math
        p(f(\cdot) \mid y) \propto p(y \mid f(\cdot))p(f(\cdot)),
        ```
        where $`p(y | f(\cdot))`$ is the likelihood and $`p(f(\cdot))`$ is the prior.

        Example:
        ```python
            >>> import gpjax as gpx
            >>>
            >>> meanf = gpx.mean_functions.Zero()
            >>> kernel = gpx.kernels.RBF()
            >>> prior = gpx.Prior(mean_function=meanf, kernel = kernel)
            >>> likelihood = gpx.likelihoods.Gaussian(num_datapoints=100)
            >>>
            >>> prior * likelihood
        ```
        Args:
            other (Likelihood): The likelihood distribution of the observed dataset.

        Returns
        -------
            Posterior: The relevant GP posterior for the given prior and
                likelihood. Special cases are accounted for where the model
                is conjugate.
        """
        return construct_posterior(prior=self, likelihood=other)

    # @overload
    # def __rmul__(self, other: Gaussian) -> "ConjugatePosterior":
    #     ...

    # @overload
    # def __rmul__(self, other: NonGaussianLikelihood) -> "NonConjugatePosterior":
    #     ...

    # @overload
    # def __rmul__(self, other: AbstractLikelihood) -> "AbstractPosterior":
    #     ...

    def __rmul__(self, other):
        r"""Combine the prior with a likelihood to form a posterior distribution.

        Reimplement the multiplication operator to allow for order-invariant
        product of a likelihood and a prior i.e., likelihood * prior.

        Args:
            other (Likelihood): The likelihood distribution of the observed
                dataset.

        Returns
        -------
            Posterior: The relevant GP posterior for the given prior and
                likelihood. Special cases are accounted for where the model
                is conjugate.
        """
        return self.__mul__(other)

    def predict(self, test_inputs: Num[Array, "N D"]) -> ReshapedGaussianDistribution:
        r"""Compute the predictive prior distribution for a given set of
        parameters. The output of this function is a function that computes
        a TFP distribution for a given set of inputs.

        In the following example, we compute the predictive prior distribution
        and then evaluate it on the interval :math:`[0, 1]`:

        Example:
        ```python
            >>> import gpjax as gpx
            >>> import jax.numpy as jnp
            >>>
            >>> kernel = gpx.kernels.RBF()
            >>> meanf = gpx.mean_functions.Zero()
            >>> prior = gpx.Prior(mean_function=meanf, kernel = kernel)
            >>>
            >>> prior.predict(jnp.linspace(0, 1, 100))
        ```

        Args:
            test_inputs (Float[Array, "N D"]): The inputs at which to evaluate the
                prior distribution.

        Returns
        -------
            ReshapedGaussianDistribution: A multivariate normal random variable representation
                of the Gaussian process, possibly with reshaped events.
        """
        x = test_inputs
        mx = jnp.atleast_1d(self.mean_function(x))
        Kxx = self.kernel.gram(x)
        Kyy = self.out_kernel.gram(jnp.arange(mx.shape[1])[:, jnp.newaxis])
        Sigma = cola.ops.Kronecker(Kxx, Kyy)
        Sigma += cola.ops.I_like(Sigma) * self.jitter

        prior_distr = GaussianDistribution(mx.flatten(), Sigma)
        if mx.shape[1] == 1:
            return prior_distr
        else:
            return ReshapedDistribution(prior_distr, mx.shape)

    def sample_approx(
        self,
        num_samples: int,
        key: KeyArray,
        num_features: Optional[int] = 100,
    ) -> FunctionalSample:
        r"""Approximate samples from the Gaussian process prior.

        Build an approximate sample from the Gaussian process prior. This method
        provides a function that returns the evaluations of a sample across any
        given inputs.

        In particular, we approximate the Gaussian processes' prior as the
        finite feature approximation
        $`\hat{f}(x) = \sum_{i=1}^m\phi_i(x)\theta_i`$ where $`\phi_i`$ are $`m`$ features
        sampled from the Fourier feature decomposition of the model's kernel and
        $`\theta_i`$ are samples from a unit Gaussian.

        A key property of such functional samples is that the same sample draw is
        evaluated for all queries. Consistency is a property that is prohibitively costly
        to ensure when sampling exactly from the GP prior, as the cost of exact sampling
        scales cubically with the size of the sample. In contrast, finite feature representations
        can be evaluated with constant cost regardless of the required number of queries.

        In the following example, we build 10 such samples and then evaluate them
        over the interval $`[0, 1]`$:

        For a `prior` distribution, the following code snippet will
        build and evaluate an approximate sample.

        Example:
        ```python
            >>> import gpjax as gpx
            >>> import jax.numpy as jnp
            >>> import jax.random as jr
            >>> key = jr.PRNGKey(123)
            >>>
            >>> meanf = gpx.mean_functions.Zero()
            >>> kernel = gpx.kernels.RBF()
            >>> prior = gpx.Prior(mean_function=meanf, kernel = kernel)
            >>>
            >>> sample_fn = prior.sample_approx(10, key)
            >>> sample_fn(jnp.linspace(0, 1, 100).reshape(-1, 1))
        ```

        Args:
            num_samples (int): The desired number of samples.
            key (KeyArray): The random seed used for the sample(s).
            num_features (int): The number of features used when approximating the
                kernel.

        Returns
        -------
            FunctionalSample: A function representing an approximate sample from the
                Gaussian process prior.
        """

        if (not isinstance(num_samples, int)) or num_samples <= 0:
            raise ValueError("num_samples must be a positive integer")

        # sample fourier features
        fourier_feature_fn = _build_fourier_features_fn(self, num_features, key)

        # sample fourier weights
        feature_weights = normal(key, [num_samples, 2 * num_features])  # [B, L]

        def sample_fn(test_inputs: Float[Array, "N D"]) -> Float[Array, "N B"]:
            feature_evals = fourier_feature_fn(test_inputs)  # [N, L]
            evaluated_sample = jnp.inner(feature_evals, feature_weights)  # [N, B]
            return self.mean_function(test_inputs) + evaluated_sample

        return sample_fn


#######################
# GP Posteriors
#######################
@dataclass
class AbstractPosterior(Module):
    r"""Abstract Gaussian process posterior.

    The base GP posterior object conditioned on an observed dataset. All
    posterior objects should inherit from this class.
    """

    prior: AbstractPrior
    likelihood: AbstractLikelihood
    jitter: float = static_field(1e-6)

    def __call__(self, *args: Any, **kwargs: Any) -> ReshapedGaussianDistribution:
        r"""Evaluate the Gaussian process posterior at the given points.

        The output of this function is a
        [TFP distribution](https://www.tensorflow.org/probability/api_docs/python/tfp/substrates/jax/distributions)
        from which the the latent function's mean and covariance can be
        evaluated and the distribution can be sampled.

        Under the hood, `__call__` is calling the objects `predict` method. For this
        reasons, classes inheriting the `AbstractPrior` class, should not overwrite the
        `__call__` method and should instead define a `predict` method.

        Args:
            *args (Any): The arguments to pass to the GP's `predict` method.
            **kwargs (Any): The keyword arguments to pass to the GP's `predict` method.

        Returns
        -------
            ReshapedGaussianDistribution: A multivariate normal random variable representation
                of the Gaussian process, possibly with reshaped events.
        """
        return self.predict(*args, **kwargs)

    @abstractmethod
    def predict(self, *args: Any, **kwargs: Any) -> ReshapedGaussianDistribution:
        r"""Compute the latent function's multivariate normal distribution for a
        given set of parameters. For any class inheriting the `AbstractPrior` class,
        this method must be implemented.

        Args:
            *args (Any): Arguments to the predict method.
            **kwargs (Any): Keyword arguments to the predict method.

        Returns
        -------
            ReshapedGaussianDistribution: A multivariate normal random variable representation
                of the Gaussian process, possibly with reshaped events.
        """
        raise NotImplementedError


@dataclass
class ConjugatePosterior(AbstractPosterior):
    r"""A Conjuate Gaussian process posterior object.

    A Gaussian process posterior distribution when the constituent likelihood
    function is a Gaussian distribution. In such cases, the latent function values
    $`f`$ can be analytically integrated out of the posterior distribution.
    As such, many computational operations can be simplified; something we make use
    of in this object.

    For a Gaussian process prior $`p(\mathbf{f})`$ and a Gaussian likelihood
    $`p(y | \mathbf{f}) = \mathcal{N}(y\mid \mathbf{f}, \sigma^2))`$ where
    $`\mathbf{f} = f(\mathbf{x})`$, the predictive posterior distribution at
    a set of inputs $`\mathbf{x}`$ is given by
    ```math
    \begin{align}
    p(\mathbf{f}^{\star}\mid \mathbf{y}) & = \int p(\mathbf{f}^{\star}, \mathbf{f} \mid \mathbf{y})\\
        & =\mathcal{N}(\mathbf{f}^{\star} \boldsymbol{\mu}_{\mid \mathbf{y}}, \boldsymbol{\Sigma}_{\mid \mathbf{y}}
    \end{align}
    ```
    where
    ```math
    \begin{align}
    \boldsymbol{\mu}_{\mid \mathbf{y}} & = k(\mathbf{x}^{\star}, \mathbf{x})\left(k(\mathbf{x}, \mathbf{x}')+\sigma^2\mathbf{I}_n\right)^{-1}\mathbf{y}  \\
    \boldsymbol{\Sigma}_{\mid \mathbf{y}} & =k(\mathbf{x}^{\star}, \mathbf{x}^{\star\prime}) -k(\mathbf{x}^{\star}, \mathbf{x})\left( k(\mathbf{x}, \mathbf{x}') + \sigma^2\mathbf{I}_n \right)^{-1}k(\mathbf{x}, \mathbf{x}^{\star}).
    \end{align}
    ```

    Example:
        ```python
            >>> import gpjax as gpx
            >>> import jax.numpy as jnp

            >>> prior = gpx.Prior(
                    mean_function = gpx.mean_functions.Zero(),
                    kernel = gpx.kernels.RBF()
                )
            >>> likelihood = gpx.likelihoods.Gaussian(num_datapoints=100)
            >>>
            >>> posterior = prior * likelihood
        ```
    """

    def predict(
        self,
        test_inputs: Num[Array, "N D"],
        train_data: Dataset,
    ) -> ReshapedGaussianDistribution:
        r"""Query the predictive posterior distribution.

        Conditional on a training data set, compute the GP's posterior
        predictive distribution for a given set of parameters. The returned function
        can be evaluated at a set of test inputs to compute the corresponding
        predictive density.

        The predictive distribution of a conjugate GP is given by
        $$
            p(\mathbf{f}^{\star}\mid \mathbf{y}) & = \int p(\mathbf{f}^{\star} \mathbf{f} \mid \mathbf{y})\\
            & =\mathcal{N}(\mathbf{f}^{\star} \boldsymbol{\mu}_{\mid \mathbf{y}}, \boldsymbol{\Sigma}_{\mid \mathbf{y}}
        $$
        where
        $$
            \boldsymbol{\mu}_{\mid \mathbf{y}} & = k(\mathbf{x}^{\star}, \mathbf{x})\left(k(\mathbf{x}, \mathbf{x}')+\sigma^2\mathbf{I}_n\right)^{-1}\mathbf{y}  \\
            \boldsymbol{\Sigma}_{\mid \mathbf{y}} & =k(\mathbf{x}^{\star}, \mathbf{x}^{\star\prime}) -k(\mathbf{x}^{\star}, \mathbf{x})\left( k(\mathbf{x}, \mathbf{x}') + \sigma^2\mathbf{I}_n \right)^{-1}k(\mathbf{x}, \mathbf{x}^{\star}).
        $$

        The conditioning set is a GPJax `Dataset` object, whilst predictions
        are made on a regular Jax array.

        Example:
            For a `posterior` distribution, the following code snippet will
            evaluate the predictive distribution.
            ```python
                >>> import gpjax as gpx
                >>> import jax.numpy as jnp
                >>>
                >>> xtrain = jnp.linspace(0, 1).reshape(-1, 1)
                >>> ytrain = jnp.sin(xtrain)
                >>> D = gpx.Dataset(X=xtrain, y=ytrain)
                >>> xtest = jnp.linspace(0, 1).reshape(-1, 1)
                >>>
                >>> prior = gpx.Prior(mean_function = gpx.Zero(), kernel = gpx.RBF())
                >>> posterior = prior * gpx.Gaussian(num_datapoints = D.n)
                >>> predictive_dist = posterior(xtest, D)
            ```

        Args:
            test_inputs (Num[Array, "N D"]): A Jax array of test inputs at which the
                predictive distribution is evaluated.
            train_data (Dataset): A `gpx.Dataset` object that contains the input and
                output data used for training dataset.

        Returns
        -------
            ReshapedGaussianDistribution: A
                function that accepts an input array and returns the predictive
                    distribution as a `GaussianDistribution` or a `ReshapedDistribution[GaussianDistribution]`.
        """
        # Unpack training data
        x, y, n_train, mask = train_data.X, train_data.y, train_data.n, train_data.mask
        m = y.shape[1]
        if m > 1 and mask is not None:
            mask = mask.flatten()
        # Unpack test inputs
        t = test_inputs
        n_test = len(test_inputs)

        # Observation noise o²
        obs_var = self.likelihood.obs_stddev**2
        mx = self.prior.mean_function(x)

        # Precompute Gram matrix, Kxx, at training inputs, x
        Kxx = self.prior.kernel.gram(x)
        Kyy = self.prior.out_kernel.gram(jnp.arange(m)[:, jnp.newaxis])

        # Σ = Kxx + Io²
        Sigma = cola.ops.Kronecker(Kxx, Kyy)
        Sigma += cola.ops.I_like(Sigma) * (obs_var + self.jitter)
        Sigma = cola.PSD(Sigma)

        if mask is not None:
            y = jnp.where(mask, 0.0, y)
            mx = jnp.where(mask, 0.0, mx)
            Sigma_masked = jnp.where(mask + mask.T, 0.0, Sigma.to_dense())
            Sigma = cola.PSD(
                Dense(
                    jnp.where(
                        jnp.diag(jnp.squeeze(mask)), 1 / (2 * jnp.pi), Sigma_masked
                    )
                )
            )

        mean_t = self.prior.mean_function(t)
        Ktt = cola.ops.Kronecker(self.prior.kernel.gram(t), Kyy)
        Ktt = cola.PSD(Ktt)
        Kxt = cola.ops.Kronecker(self.prior.kernel.cross_covariance(x, t), Kyy)

        # Σ⁻¹ Kxt
        if mask is not None:
            Kxt = jnp.where(mask * jnp.ones((1, n_train), dtype=bool), 0.0, Kxt)
        Sigma_inv_Kxt = cola.solve(Sigma, Kxt, Cholesky())

        # μt  +  Ktx (Kxx + Io²)⁻¹ (y  -  μx)
        mean = mean_t.flatten() + Sigma_inv_Kxt.T @ (y - mx).flatten()

        # Ktt  -  Ktx (Kxx + Io²)⁻¹ Kxt, TODO: Take advantage of covariance structure to compute Schur complement more efficiently.
        covariance = Ktt - Kxt.T @ Sigma_inv_Kxt
        covariance += cola.ops.I_like(covariance) * self.prior.jitter
        covariance = cola.PSD(covariance)
        rval = GaussianDistribution(jnp.atleast_1d(mean.squeeze()), covariance)
        if m == 1:
            return rval
        else:
            return ReshapedDistribution(rval, (n_test, m))

    def sample_approx(
        self,
        num_samples: int,
        train_data: Dataset,
        key: KeyArray,
        num_features: Optional[int] = 100,
    ) -> FunctionalSample:
        r"""Draw approximate samples from the Gaussian process posterior.

        Build an approximate sample from the Gaussian process posterior. This method
        provides a function that returns the evaluations of a sample across any given
        inputs.

        Unlike when building approximate samples from a Gaussian process prior, decompositions
        based on Fourier features alone rarely give accurate samples. Therefore, we must also
        include an additional set of features (known as canonical features) to better model the
        transition from Gaussian process prior to Gaussian process posterior. For more details
        see [Wilson et. al. (2020)](https://arxiv.org/abs/2002.09309).

        In particular, we approximate the Gaussian processes' posterior as the finite
        feature approximation
        $`\hat{f}(x) = \sum_{i=1}^m \phi_i(x)\theta_i + \sum{j=1}^N v_jk(.,x_j)`$
        where $`\phi_i`$ are m features sampled from the Fourier feature decomposition of
        the model's kernel and $`k(., x_j)`$ are N canonical features. The Fourier
        weights $`\theta_i`$ are samples from a unit Gaussian. See
        [Wilson et. al. (2020)](https://arxiv.org/abs/2002.09309) for expressions
        for the canonical weights $`v_j`$.

        A key property of such functional samples is that the same sample draw is
        evaluated for all queries. Consistency is a property that is prohibitively costly
        to ensure when sampling exactly from the GP prior, as the cost of exact sampling
        scales cubically with the size of the sample. In contrast, finite feature representations
        can be evaluated with constant cost regardless of the required number of queries.

        Args:
            num_samples (int): The desired number of samples.
            key (KeyArray): The random seed used for the sample(s).
            num_features (int): The number of features used when approximating the
                kernel.

        Returns
        -------
            FunctionalSample: A function representing an approximate sample from the Gaussian
            process prior.
        """
        if (not isinstance(num_samples, int)) or num_samples <= 0:
            raise ValueError("num_samples must be a positive integer")

        # sample fourier features
        fourier_feature_fn = _build_fourier_features_fn(self.prior, num_features, key)

        # sample fourier weights
        fourier_weights = normal(key, [num_samples, 2 * num_features])  # [B, L]

        # sample weights v for canonical features
        # v = Σ⁻¹ (y + ε - ɸ⍵) for  Σ = Kxx + Io² and ε ᯈ N(0, o²)
        obs_var = self.likelihood.obs_stddev**2
        Kxx = self.prior.kernel.gram(train_data.X)  #  [N, N]
        Sigma = Kxx + cola.ops.I_like(Kxx) * (obs_var + self.jitter)  #  [N, N]
        eps = jnp.sqrt(obs_var) * normal(key, [train_data.n, num_samples])  #  [N, B]
        y = train_data.y - self.prior.mean_function(train_data.X)  # account for mean
        Phi = fourier_feature_fn(train_data.X)
        canonical_weights = cola.solve(
            Sigma,
            y + eps - jnp.inner(Phi, fourier_weights),
            Cholesky(),
        )  #  [N, B]

        def sample_fn(test_inputs: Float[Array, "n D"]) -> Float[Array, "n B"]:
            fourier_features = fourier_feature_fn(test_inputs)  # [n, L]
            weight_space_contribution = jnp.inner(
                fourier_features, fourier_weights
            )  # [n, B]
            canonical_features = self.prior.kernel.cross_covariance(
                test_inputs, train_data.X
            )  # [n, N]
            function_space_contribution = jnp.matmul(
                canonical_features, canonical_weights
            )

            return (
                self.prior.mean_function(test_inputs)
                + weight_space_contribution
                + function_space_contribution
            )

        return sample_fn


@dataclass
class NonConjugatePosterior(AbstractPosterior):
    r"""A non-conjugate Gaussian process posterior object.

    A Gaussian process posterior object for models where the likelihood is
    non-Gaussian. Unlike the `ConjugatePosterior` object, the
    `NonConjugatePosterior` object does not provide an exact marginal
    log-likelihood function. Instead, the `NonConjugatePosterior` object
    represents the posterior distributions as a function of the model's
    hyperparameters and the latent function. Markov chain Monte Carlo,
    variational inference, or Laplace approximations can then be used to sample
    from, or optimise an approximation to, the posterior distribution.
    """

    latent: Union[Float[Array, "N 1"], None] = param_field(None)
    key: KeyArray = static_field(PRNGKey(42))

    def __post_init__(self):
        if self.latent is None:
            self.latent = normal(self.key, shape=(self.likelihood.num_datapoints, 1))

    def predict(
        self, test_inputs: Num[Array, "N D"], train_data: Dataset
    ) -> GaussianDistribution:
        r"""Query the predictive posterior distribution.

        Conditional on a set of training data, compute the GP's posterior
        predictive distribution for a given set of parameters. The returned
        function can be evaluated at a set of test inputs to compute the
        corresponding predictive density. Note, to gain predictions on the scale
        of the original data, the returned distribution will need to be
        transformed through the likelihood function's inverse link function.

        Args:
            train_data (Dataset): A `gpx.Dataset` object that contains the input
                and output data used for training dataset.

        Returns
        -------
            GaussianDistribution: A function that accepts an
                input array and returns the predictive distribution as
                a `dx.Distribution`.
        """
        # Unpack training data
        x = train_data.X

        # Unpack mean function and kernel
        mean_function = self.prior.mean_function
        kernel = self.prior.kernel

        # Precompute lower triangular of Gram matrix, Lx, at training inputs, x
        Kxx = kernel.gram(x)
        Kxx += cola.ops.I_like(Kxx) * self.prior.jitter
        Kxx = cola.PSD(Kxx)
        Lx = lower_cholesky(Kxx)

        # Unpack test inputs
        t = test_inputs

        # Compute terms of the posterior predictive distribution
        Ktx = kernel.cross_covariance(t, x)
        Ktt = kernel.gram(t)
        mean_t = mean_function(t)

        # Lx⁻¹ Kxt
        Lx_inv_Kxt = cola.solve(Lx, Ktx.T, Cholesky())

        # Whitened function values, wx, corresponding to the inputs, x
        wx = self.latent

        # μt + Ktx Lx⁻¹ wx
        mean = mean_t + jnp.matmul(Lx_inv_Kxt.T, wx)

        # Ktt - Ktx Kxx⁻¹ Kxt, TODO: Take advantage of covariance structure to compute Schur complement more efficiently.
        covariance = Ktt - jnp.matmul(Lx_inv_Kxt.T, Lx_inv_Kxt)
        covariance += cola.ops.I_like(covariance) * self.prior.jitter
        covariance = cola.PSD(covariance)

        return GaussianDistribution(jnp.atleast_1d(mean.squeeze()), covariance)


#######################
# Utils
#######################


@overload
def construct_posterior(prior: Prior, likelihood: Gaussian) -> ConjugatePosterior:
    ...


@overload
def construct_posterior(
    prior: Prior, likelihood: NonGaussianLikelihood
) -> NonConjugatePosterior:
    ...


@overload
def construct_posterior(
    prior: Prior, likelihood: AbstractLikelihood
) -> AbstractPosterior:
    ...


def construct_posterior(prior, likelihood):
    r"""Utility function for constructing a posterior object from a prior and
    likelihood. The function will automatically select the correct posterior
    object based on the likelihood.

    Args:
        prior (Prior): The Prior distribution.
        likelihood (AbstractLikelihood): The likelihood that represents our
            beliefs around the distribution of the data.

    Returns
    -------
        AbstractPosterior: A posterior distribution. If the likelihood is
            Gaussian, then a `ConjugatePosterior` will be returned. Otherwise,
            a `NonConjugatePosterior` will be returned.
    """
    if isinstance(likelihood, Gaussian):
        return ConjugatePosterior(prior=prior, likelihood=likelihood)

    return NonConjugatePosterior(prior=prior, likelihood=likelihood)


def _build_fourier_features_fn(
    prior: Prior, num_features: int, key: KeyArray
) -> Callable[[Float[Array, "N D"]], Float[Array, "N L"]]:
    r"""Return a function that evaluates features sampled from the Fourier feature
    decomposition of the prior's kernel.

    Args:
        prior (Prior): The Prior distribution.
        num_features (int): The number of feature functions to be sampled.
        key (KeyArray): The random seed used.

    Returns
    -------
        Callable: A callable function evaluation the sampled feature functions.
    """
    if (not isinstance(num_features, int)) or num_features <= 0:
        raise ValueError("num_features must be a positive integer")

    # Approximate kernel with feature decomposition
    approximate_kernel = RFF(
        base_kernel=prior.kernel, num_basis_fns=num_features, key=key
    )

    def eval_fourier_features(test_inputs: Float[Array, "N D"]) -> Float[Array, "N L"]:
        Phi = approximate_kernel.compute_features(x=test_inputs)
        Phi *= jnp.sqrt(prior.kernel.variance / num_features)
        return Phi

    return eval_fourier_features


__all__ = [
    "AbstractPrior",
    "Prior",
    "AbstractPosterior",
    "ConjugatePosterior",
    "NonConjugatePosterior",
    "construct_posterior",
]



# Copyright 2023 The GPJax Contributors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from abc import abstractmethod
from dataclasses import dataclass

import jax.numpy as jnp
import jax.random as jr
from jaxtyping import (
    Array,
    Float,
    Integer,
)

from gpjax.dataset import Dataset
from gpjax.decision_making.search_space import ContinuousSearchSpace
from gpjax.typing import KeyArray


@dataclass
class PoissonTestFunction:
    """
    Test function for GPs utilising the Poisson likelihood. Function taken from
    https://docs.jaxgaussianprocesses.com/examples/poisson/#dataset.

    Attributes:
        search_space (ContinuousSearchSpace): Search space for the function.
    """

    search_space = ContinuousSearchSpace(
        lower_bounds=jnp.array([-2.0]),
        upper_bounds=jnp.array([2.0]),
    )

    def generate_dataset(self, num_points: int, key: KeyArray) -> Dataset:
        """
        Generate a toy dataset from the test function.

        Args:
            num_points (int): Number of points to sample.
            key (KeyArray): JAX PRNG key.

        Returns:
            Dataset: Dataset of points sampled from the test function.
        """
        X = self.search_space.sample(num_points=num_points, key=key)
        y = self.evaluate(X)
        return Dataset(X=X, y=y)

    def generate_test_points(
        self, num_points: int, key: KeyArray
    ) -> Float[Array, "N D"]:
        """
        Generate test points from the search space of the test function.

        Args:
            num_points (int): Number of points to sample.
            key (KeyArray): JAX PRNG key.

        Returns:
            Float[Array, 'N D']: Test points sampled from the search space.
        """
        return self.search_space.sample(num_points=num_points, key=key)

    @abstractmethod
    def evaluate(self, x: Float[Array, "N 1"]) -> Integer[Array, "N 1"]:
        """
        Evaluate the test function at a set of points. Function taken from
        https://docs.jaxgaussianprocesses.com/examples/poisson/#dataset.

        Args:
            x (Float[Array, 'N D']): Points to evaluate the test function at.

        Returns:
            Integer[Array, 'N 1']: Values of the test function at the points.
        """
        key = jr.PRNGKey(42)
        f = lambda x: 2.0 * jnp.sin(3 * x) + 0.5 * x
        return jr.poisson(key, jnp.exp(f(x)))



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import abc
from dataclasses import dataclass

from beartype.typing import (
    Any,
    Union,
)
from jax import vmap
import jax.numpy as jnp
import jax.scipy as jsp
from jaxtyping import Float
import tensorflow_probability.substrates.jax as tfp

from gpjax.base import (
    Module,
    param_field,
    static_field,
)
from gpjax.distributions import (
    GaussianDistribution,
    ReshapedDistribution,
)
from gpjax.integrators import (
    AbstractIntegrator,
    AnalyticalGaussianIntegrator,
    GHQuadratureIntegrator,
)
from gpjax.typing import (
    Array,
    ScalarFloat,
)

tfb = tfp.bijectors
tfd = tfp.distributions


@dataclass
class AbstractLikelihood(Module):
    r"""Abstract base class for likelihoods."""

    num_datapoints: int = static_field()
    integrator: AbstractIntegrator = static_field(GHQuadratureIntegrator())

    def __call__(self, *args: Any, **kwargs: Any) -> tfd.Distribution:
        r"""Evaluate the likelihood function at a given predictive distribution.

        Args:
            *args (Any): Arguments to be passed to the likelihood's `predict` method.
            **kwargs (Any): Keyword arguments to be passed to the likelihood's
                `predict` method.

        Returns
        -------
            tfd.Distribution: The predictive distribution.
        """
        return self.predict(*args, **kwargs)

    @abc.abstractmethod
    def predict(self, *args: Any, **kwargs: Any) -> tfd.Distribution:
        r"""Evaluate the likelihood function at a given predictive distribution.

        Args:
            *args (Any): Arguments to be passed to the likelihood's `predict` method.
            **kwargs (Any): Keyword arguments to be passed to the likelihood's
                `predict` method.

        Returns
        -------
            tfd.Distribution: The predictive distribution.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def link_function(self, f: Float[Array, "..."]) -> tfd.Distribution:
        r"""Return the link function of the likelihood function.

        Returns
        -------
            tfd.Distribution: The distribution of observations, y, given values of the
                Gaussian process, f.
        """
        raise NotImplementedError

    def expected_log_likelihood(
        self,
        y: Float[Array, "N D"],
        mean: Float[Array, "N D"],
        variance: Float[Array, "N D"],
    ) -> Float[Array, " N"]:
        r"""Compute the expected log likelihood.

        For a variational distribution $`q(f)\sim\mathcal{N}(m, s)`$ and a likelihood
        $`p(y|f)`$, compute the expected log likelihood:
        ```math
        \mathbb{E}_{q(f)}\left[\log p(y|f)\right]
        ```

        Args:
            y (Float[Array, 'N D']): The observed response variable.
            mean (Float[Array, 'N D']): The variational mean.
            variance (Float[Array, 'N D']): The variational variance.

        Returns:
            ScalarFloat: The expected log likelihood.
        """
        log_prob = vmap(lambda f, y: self.link_function(f).log_prob(y))
        return self.integrator(
            fun=log_prob, y=y, mean=mean, variance=variance, likelihood=self
        )


@dataclass
class Gaussian(AbstractLikelihood):
    r"""Gaussian likelihood object.

    Args:
        obs_stddev (Union[ScalarFloat, Float[Array, "#N"]]): the standard deviation
            of the Gaussian observation noise.

    """

    obs_stddev: Union[ScalarFloat, Float[Array, "#N"]] = param_field(
        jnp.array(1.0), bijector=tfb.Softplus()
    )
    integrator: AbstractIntegrator = static_field(AnalyticalGaussianIntegrator())

    def link_function(self, f: Float[Array, "..."]) -> tfd.Normal:
        r"""The link function of the Gaussian likelihood.

        Args:
            f (Float[Array, "..."]): Function values.

        Returns
        -------
            tfd.Normal: The likelihood function.
        """
        return tfd.Normal(loc=f, scale=self.obs_stddev.astype(f.dtype))

    def predict(
        self, dist: Union[tfd.MultivariateNormalTriL, GaussianDistribution]
    ) -> tfd.MultivariateNormalFullCovariance:
        r"""Evaluate the Gaussian likelihood.

        Evaluate the Gaussian likelihood function at a given predictive
        distribution. Computationally, this is equivalent to summing the
        observation noise term to the diagonal elements of the predictive
        distribution's covariance matrix.

        Args:
            dist (tfd.Distribution): The Gaussian process posterior,
                evaluated at a finite set of test points.

        Returns
        -------
            tfd.Distribution: The predictive distribution.
        """
        mean = dist.mean()
        event_size = mean.size

        cov = dist.covariance()

        # reshape for handling multi-output case
        mean = mean.flatten()
        cov = cov.reshape([mean.size] * 2)

        noisy_cov = cov.at[jnp.diag_indices(event_size)].add(self.obs_stddev**2)

        likelihood_distr = tfd.MultivariateNormalFullCovariance(mean, noisy_cov)
        if len(dist.event_shape) == 1:
            return likelihood_distr
        else:
            return ReshapedDistribution(likelihood_distr, dist.event_shape)


@dataclass
class Bernoulli(AbstractLikelihood):
    def link_function(self, f: Float[Array, "..."]) -> tfd.Distribution:
        r"""The probit link function of the Bernoulli likelihood.

        Args:
            f (Float[Array, "..."]): Function values.

        Returns
        -------
            tfd.Distribution: The likelihood function.
        """
        return tfd.Bernoulli(probs=inv_probit(f))

    def predict(self, dist: tfd.Distribution) -> tfd.Distribution:
        r"""Evaluate the pointwise predictive distribution.

        Evaluate the pointwise predictive distribution, given a Gaussian
        process posterior and likelihood parameters.

        Args:
            dist (tfd.Distribution): The Gaussian process posterior, evaluated
                at a finite set of test points.

        Returns
        -------
            tfd.Distribution: The pointwise predictive distribution.
        """
        variance = jnp.diag(dist.covariance())
        mean = dist.mean().ravel()
        return self.link_function(mean / jnp.sqrt(1.0 + variance))


@dataclass
class Poisson(AbstractLikelihood):
    def link_function(self, f: Float[Array, "..."]) -> tfd.Distribution:
        r"""The link function of the Poisson likelihood.

        Args:
            f (Float[Array, "..."]): Function values.

        Returns:
            tfd.Distribution: The likelihood function.
        """
        return tfd.Poisson(rate=jnp.exp(f))

    def predict(self, dist: tfd.Distribution) -> tfd.Distribution:
        r"""Evaluate the pointwise predictive distribution.

        Evaluate the pointwise predictive distribution, given a Gaussian
        process posterior and likelihood parameters.

        Args:
            dist (tfd.Distribution): The Gaussian process posterior, evaluated
                at a finite set of test points.

        Returns:
            tfd.Distribution: The pointwise predictive distribution.
        """
        return self.link_function(dist.mean())


def inv_probit(x: Float[Array, " *N"]) -> Float[Array, " *N"]:
    r"""Compute the inverse probit function.

    Args:
        x (Float[Array, "*N"]): A vector of values.

    Returns
    -------
        Float[Array, "*N"]: The inverse probit of the input vector.
    """
    jitter = 1e-3  # To ensure output is in interval (0, 1).
    return 0.5 * (1.0 + jsp.special.erf(x / jnp.sqrt(2.0))) * (1 - 2 * jitter) + jitter


NonGaussianLikelihood = Union[Poisson, Bernoulli]

__all__ = [
    "AbstractLikelihood",
    "NonGaussianLikelihood",
    "Gaussian",
    "Bernoulli",
    "Poisson",
    "inv_probit",
]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['gpjax',
 'gpjax.base',
 'gpjax.decision_making',
 'gpjax.decision_making.test_functions',
 'gpjax.decision_making.utility_functions',
 'gpjax.kernels',
 'gpjax.kernels.approximations',
 'gpjax.kernels.computations',
 'gpjax.kernels.non_euclidean',
 'gpjax.kernels.nonstationary',
 'gpjax.kernels.stationary']

package_data = \
{'': ['*']}

install_requires = \
['beartype>=0.16.2,<0.17.0',
 'cola-ml>=0.0.5,<0.0.6',
 'jax>=0.4.10',
 'jaxlib>=0.4.10',
 'jaxtyping>=0.2.15,<0.3.0',
 'optax>=0.1.4,<0.2.0',
 'orbax-checkpoint>=0.2.3',
 'simple-pytree>=0.1.7,<0.2.0',
 'tensorflow-probability>=0.19.0,<0.20.0',
 'tqdm>=4.65.0,<5.0.0']

setup_kwargs = {
    'name': 'gpjax',
    'version': '0.7.2',
    'description': 'Gaussian processes in JAX.',
    'long_description': '<!-- <h1 align=\'center\'>GPJax</h1>\n<h2 align=\'center\'>Gaussian processes in Jax.</h2> -->\n<p align="center">\n<img width="700" height="300" src="https://raw.githubusercontent.com/JaxGaussianProcesses/GPJax/main/docs/_static/gpjax_logo.svg" alt="GPJax\'s logo">\n</p>\n\n[![codecov](https://codecov.io/gh/JaxGaussianProcesses/GPJax/branch/master/graph/badge.svg?token=DM1DRDASU2)](https://codecov.io/gh/JaxGaussianProcesses/GPJax)\n[![CodeFactor](https://www.codefactor.io/repository/github/jaxgaussianprocesses/gpjax/badge)](https://www.codefactor.io/repository/github/jaxgaussianprocesses/gpjax)\n[![Netlify Status](https://api.netlify.com/api/v1/badges/d3950e6f-321f-4508-9e52-426b5dae2715/deploy-status)](https://app.netlify.com/sites/endearing-crepe-c2d5fe/deploys)\n[![PyPI version](https://badge.fury.io/py/GPJax.svg)](https://badge.fury.io/py/GPJax)\n[![DOI](https://joss.theoj.org/papers/10.21105/joss.04455/status.svg)](https://doi.org/10.21105/joss.04455)\n[![Downloads](https://pepy.tech/badge/gpjax)](https://pepy.tech/project/gpjax)\n[![Slack Invite](https://img.shields.io/badge/Slack_Invite--blue?style=social&logo=slack)](https://join.slack.com/t/gpjax/shared_invite/zt-1da57pmjn-rdBCVg9kApirEEn2E5Q2Zw)\n\n[**Quickstart**](#simple-example)\n| [**Install guide**](#installation)\n| [**Documentation**](https://docs.jaxgaussianprocesses.com/)\n| [**Slack Community**](https://join.slack.com/t/gpjax/shared_invite/zt-1da57pmjn-rdBCVg9kApirEEn2E5Q2Zw)\n\nGPJax aims to provide a low-level interface to Gaussian process (GP) models in\n[Jax](https://github.com/google/jax), structured to give researchers maximum\nflexibility in extending the code to suit their own needs. The idea is that the\ncode should be as close as possible to the maths we write on paper when working\nwith GP models.\n\n# Package organisation\n\n## Contributions\n\nWe would be delighted to receive contributions from interested individuals and\ngroups. To learn how you can get involved, please read our [guide for\ncontributing](https://github.com/JaxGaussianProcesses/GPJax/blob/main/docs/contributing.md).\nIf you have any questions, we encourage you to [open an\nissue](https://github.com/JaxGaussianProcesses/GPJax/issues/new/choose). For\nbroader conversations, such as best GP fitting practices or questions about the\nmathematics of GPs, we invite you to [open a\ndiscussion](https://github.com/JaxGaussianProcesses/GPJax/discussions).\n\nAnother way you can contribute to GPJax is through [issue\ntriaging](https://www.codetriage.com/what).  This can include reproducing bug reports,\nasking for vital information such as version numbers and reproduction instructions, or\nidentifying stale issues. If you would like to begin triaging issues, an easy way to get\nstarted is to \n[subscribe to GPJax on CodeTriage](https://www.codetriage.com/jaxgaussianprocesses/gpjax).\n\nAs a contributor to GPJax, you are expected to abide by our [code of\nconduct](docs/CODE_OF_CONDUCT.md). If you are feel that you have either experienced or\nwitnessed behaviour that violates this standard, then we ask that you report any such\nbehaviours though [this form](https://jaxgaussianprocesses.com/contact/) or reach out to \none of the project\'s [_gardeners_](https://docs.jaxgaussianprocesses.com/GOVERNANCE/#roles).\n\nFeel free to join our [Slack\nChannel](https://join.slack.com/t/gpjax/shared_invite/zt-1da57pmjn-rdBCVg9kApirEEn2E5Q2Zw),\nwhere we can discuss the development of GPJax and broader support for Gaussian\nprocess modelling.\n\n\n## Governance\n\nGPJax was founded by [Thomas Pinder](https://github.com/thomaspinder). Today, the\nproject\'s gardeners are [daniel-dodd@](https://github.com/daniel-dodd),\n[henrymoss@](https://github.com/henrymoss), [st--@](https://github.com/st--), and\n[thomaspinder@](https://github.com/thomaspinder), listed in alphabetical order. The full\ngovernance structure of GPJax is detailed [here](docs/GOVERNANCE.md).We appreciate all\n[the contributors to\nGPJax](https://github.com/JaxGaussianProcesses/GPJax/graphs/contributors) who have\nhelped to shape GPJax into the package it is today.\n\n\n# Supported methods and interfaces\n\n## Notebook examples\n\n> - [**Conjugate Inference**](https://docs.jaxgaussianprocesses.com/examples/regression/)\n> - [**Classification with MCMC**](https://docs.jaxgaussianprocesses.com/examples/classification/)\n> - [**Sparse Variational Inference**](https://docs.jaxgaussianprocesses.com/examples/collapsed_vi/)\n> - [**Stochastic Variational Inference**](https://docs.jaxgaussianprocesses.com/examples/uncollapsed_vi/)\n> - [**BlackJax Integration**](https://docs.jaxgaussianprocesses.com/examples/classification/#mcmc-inference)\n> - [**Laplace Approximation**](https://docs.jaxgaussianprocesses.com/examples/classification/#laplace-approximation)\n> - [**Inference on Non-Euclidean Spaces**](https://docs.jaxgaussianprocesses.com/examples/constructing_new_kernels/#custom-kernel)\n> - [**Inference on Graphs**](https://docs.jaxgaussianprocesses.com/examples/graph_kernels/)\n> - [**Pathwise Sampling**](https://docs.jaxgaussianprocesses.com/examples/spatial/)\n> - [**Learning Gaussian Process Barycentres**](https://docs.jaxgaussianprocesses.com/examples/barycentres/)\n> - [**Deep Kernel Regression**](https://docs.jaxgaussianprocesses.com/examples/deep_kernels/)\n> - [**Poisson Regression**](https://docs.jaxgaussianprocesses.com/examples/poisson/)\n> - [**Bayesian Optimisation**](https://docs.jaxgaussianprocesses.com/examples/bayesian_optimisation/)\n\n## Guides for customisation\n>\n> - [**Custom kernels**](https://docs.jaxgaussianprocesses.com/examples/constructing_new_kernels/#custom-kernel)\n> - [**UCI regression**](https://docs.jaxgaussianprocesses.com/examples/yacht/)\n\n## Conversion between `.ipynb` and `.py`\nAbove examples are stored in [examples](docs/examples) directory in the double\npercent (`py:percent`) format. Checkout [jupytext\nusing-cli](https://jupytext.readthedocs.io/en/latest/using-cli.html) for more\ninfo.\n\n* To convert `example.py` to `example.ipynb`, run:\n\n```bash\njupytext --to notebook example.py\n```\n\n* To convert `example.ipynb` to `example.py`, run:\n\n```bash\njupytext --to py:percent example.ipynb\n```\n\n# Simple example\n\nLet us import some dependencies and simulate a toy dataset $\\mathcal{D}$.\n\n```python\nfrom jax import config\n\nconfig.update("jax_enable_x64", True)\n\nimport gpjax as gpx\nfrom jax import grad, jit\nimport jax.numpy as jnp\nimport jax.random as jr\nimport optax as ox\n\nkey = jr.PRNGKey(123)\n\nf = lambda x: 10 * jnp.sin(x)\n\nn = 50\nx = jr.uniform(key=key, minval=-3.0, maxval=3.0, shape=(n,1)).sort()\ny = f(x) + jr.normal(key, shape=(n,1))\nD = gpx.Dataset(X=x, y=y)\n\n# Construct the prior\nmeanf = gpx.mean_functions.Zero()\nkernel = gpx.kernels.RBF()\nprior = gpx.Prior(mean_function=meanf, kernel = kernel)\n\n# Define a likelihood\nlikelihood = gpx.Gaussian(num_datapoints = n)\n\n# Construct the posterior\nposterior = prior * likelihood\n\n# Define an optimiser\noptimiser = ox.adam(learning_rate=1e-2)\n\n# Define the marginal log-likelihood\nnegative_mll = jit(gpx.objectives.ConjugateMLL(negative=True))\n\n# Obtain Type 2 MLEs of the hyperparameters\nopt_posterior, history = gpx.fit(\n    model=posterior,\n    objective=negative_mll,\n    train_data=D,\n    optim=optimiser,\n    num_iters=500,\n    safe=True,\n    key=key,\n)\n\n# Infer the predictive posterior distribution\nxtest = jnp.linspace(-3., 3., 100).reshape(-1, 1)\nlatent_dist = opt_posterior(xtest, D)\npredictive_dist = opt_posterior.likelihood(latent_dist)\n\n# Obtain the predictive mean and standard deviation\npred_mean = predictive_dist.mean()\npred_std = predictive_dist.stddev()\n```\n\n# Installation\n\n## Stable version\n\nThe latest stable version of GPJax can be installed via\npip:\n\n```bash\npip install gpjax\n```\n\n> **Note**\n>\n> We recommend you check your installation version:\n> ```python\n> python -c \'import gpjax; print(gpjax.__version__)\'\n> ```\n\n\n\n## Development version\n> **Warning**\n>\n> This version is possibly unstable and may contain bugs.\n\n> **Note**\n>\n> We advise you create virtual environment before installing:\n> ```\n> conda create -n gpjax_experimental python=3.10.0\n> conda activate gpjax_experimental\n>  ```\n\n\nClone a copy of the repository to your local machine and run the setup\nconfiguration in development mode.\n```bash\ngit clone https://github.com/JaxGaussianProcesses/GPJax.git\ncd GPJax\npoetry install\n```\n\n> We recommend you check your installation passes the supplied unit tests:\n>\n> ```python\n> poetry run pytest\n> ```\n\n# Citing GPJax\n\nIf you use GPJax in your research, please cite our [JOSS paper](https://joss.theoj.org/papers/10.21105/joss.04455#).\n\n```\n@article{Pinder2022,\n  doi = {10.21105/joss.04455},\n  url = {https://doi.org/10.21105/joss.04455},\n  year = {2022},\n  publisher = {The Open Journal},\n  volume = {7},\n  number = {75},\n  pages = {4455},\n  author = {Thomas Pinder and Daniel Dodd},\n  title = {GPJax: A Gaussian Process Framework in JAX},\n  journal = {Journal of Open Source Software}\n}\n```\n',
    'author': 'Thomas Pinder',
    'author_email': 'tompinder@live.co.uk',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/JaxGaussianProcesses/GPJax',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<3.12',
}


setup(**setup_kwargs)



# Copyright 2022 The JaxGaussianProcesses Contributors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

from gpjax.base.module import (
    Module,
    load_tree,
    meta,
    meta_flatten,
    meta_leaves,
    meta_map,
    save_tree,
    static_field,
)
from gpjax.base.param import param_field

__all__ = [
    "Module",
    "meta_leaves",
    "meta_flatten",
    "meta_map",
    "meta",
    "param_field",
    "static_field",
    "save_tree",
    "load_tree",
]



# Copyright 2023 The GPJax Contributors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from abc import (
    ABC,
    abstractmethod,
)
import copy
from dataclasses import dataclass

from beartype.typing import (
    Callable,
    Dict,
    List,
    Mapping,
)
import jax.numpy as jnp
import jax.random as jr

from gpjax.dataset import Dataset
from gpjax.decision_making.posterior_handler import PosteriorHandler
from gpjax.decision_making.search_space import AbstractSearchSpace
from gpjax.decision_making.utility_functions import (
    AbstractUtilityFunctionBuilder,
    ThompsonSampling,
)
from gpjax.decision_making.utility_maximizer import AbstractUtilityMaximizer
from gpjax.decision_making.utils import FunctionEvaluator
from gpjax.gps import AbstractPosterior
from gpjax.typing import (
    Array,
    Float,
    KeyArray,
)


@dataclass
class AbstractDecisionMaker(ABC):
    """
    AbstractDecisionMaker abstract base class which handles the core decision making
    loop, where we sequentially decide on points to query our function of interest at.
    The decision making loop is split into two key steps, `ask` and `tell`. The `ask`
    step is typically used to decide which point to query next. The `tell` step is
    typically used to update models and datasets with newly queried points. These steps
    can be combined in a 'run' loop which alternates between asking which point to query
    next and telling the decision maker about the newly queried point having evaluated
    the black-box function of interest at this point.

    Attributes:
        search_space (AbstractSearchSpace): Search space over which we can evaluate the
        function(s) of interest.
        posterior_handlers (Dict[str, PosteriorHandler]): Dictionary of posterior
            handlers, which are used to update posteriors throughout the decision making
            loop. Note that the word `posteriors` is used for consistency with GPJax, but these
            objects are typically referred to as `models` in the model-based decision
            making literature. Tags are used to distinguish between posteriors. In a typical
            Bayesian optimisation setup one of the tags will be `OBJECTIVE`, defined in
            decision_making.utils.
        datasets (Dict[str, Dataset]): Dictionary of datasets, which are augmented with
            observations throughout the decision making loop. In a typical setup they are
            also used to update the posteriors, using the `posterior_handlers`. Tags are used
            to distinguish datasets, and correspond to tags in `posterior_handlers`.
        key (KeyArray): JAX random key, used to generate random numbers.
        batch_size (int): Number of points to query at each step of the decision making
            loop. Note that `SinglePointUtilityFunction`s are only capable of generating
            one point to be queried at each iteration of the decision making loop.
        post_ask (List[Callable]): List of functions to be executed after each ask step.
        post_tell (List[Callable]): List of functions to be executed after each tell
            step.
    """

    search_space: AbstractSearchSpace
    posterior_handlers: Dict[str, PosteriorHandler]
    datasets: Dict[str, Dataset]
    key: KeyArray
    batch_size: int
    post_ask: List[
        Callable
    ]  # Specific type is List[Callable[[AbstractDecisionMaker, Float[Array, ["B D"]]], None]] but causes Beartype issues
    post_tell: List[
        Callable
    ]  # Specific type is List[Callable[[AbstractDecisionMaker], None]] but causes Beartype issues

    def __post_init__(self):
        """
        At initialisation we check that the posterior handlers and datasets are
        consistent (i.e. have the same tags), and then initialise the posteriors, optimizing them using the
        corresponding datasets.
        """
        self.datasets = copy.copy(
            self.datasets
        )  # Ensure initial datasets passed in to DecisionMaker are not mutated from within

        if self.batch_size < 1:
            raise ValueError(
                f"Batch size must be greater than 0, got {self.batch_size}."
            )

        # Check that posterior handlers and datasets are consistent
        if self.posterior_handlers.keys() != self.datasets.keys():
            raise ValueError(
                "Posterior handlers and datasets must have the same keys. "
                f"Got posterior handlers keys {self.posterior_handlers.keys()} and "
                f"datasets keys {self.datasets.keys()}."
            )

        # Initialize posteriors
        self.posteriors: Dict[str, AbstractPosterior] = {}
        for tag, posterior_handler in self.posterior_handlers.items():
            self.posteriors[tag] = posterior_handler.get_posterior(
                self.datasets[tag], optimize=True, key=self.key
            )

    @abstractmethod
    def ask(self, key: KeyArray) -> Float[Array, "B D"]:
        """
        Get the point(s) to be queried next.

        Args:
            key (KeyArray): JAX PRNG key for controlling random state.

        Returns:
            Float[Array, "1 D"]: Point to be queried next
        """
        raise NotImplementedError

    def tell(self, observation_datasets: Mapping[str, Dataset], key: KeyArray):
        """
        Add newly observed data to datasets and update the corresponding posteriors.

        Args:
            observation_datasets (Mapping[str, Dataset]): Dictionary of datasets
            containing new observations. Tags are used to distinguish datasets, and
            correspond to tags in `posterior_handlers` and `self.datasets`.
            key (KeyArray): JAX PRNG key for controlling random state.
        """
        if observation_datasets.keys() != self.datasets.keys():
            raise ValueError(
                "Observation datasets and existing datasets must have the same keys. "
                f"Got observation datasets keys {observation_datasets.keys()} and "
                f"existing datasets keys {self.datasets.keys()}."
            )

        for tag, observation_dataset in observation_datasets.items():
            self.datasets[tag] += observation_dataset

        for tag, posterior_handler in self.posterior_handlers.items():
            key, _ = jr.split(key)
            self.posteriors[tag] = posterior_handler.update_posterior(
                self.datasets[tag], self.posteriors[tag], optimize=True, key=key
            )

    def run(
        self, n_steps: int, black_box_function_evaluator: FunctionEvaluator
    ) -> Mapping[str, Dataset]:
        """
        Run the decision making loop continuously for for `n_steps`. This is broken down
        into three main steps:
        1. Call the `ask` method to get the point to be queried next.
        2. Call the `black_box_function_evaluator` to evaluate the black box functions
        of interest at the point chosen to be queried.
        3. Call the `tell` method to update the datasets and posteriors with the newly
        observed data.

        In addition to this, after the `ask` step, the functions in the `post_ask` list
        are executed, taking as arguments the decision maker and the point chosen to be
        queried next. Similarly, after the `tell` step, the functions in the `post_tell`
        list are executed, taking the decision maker as the sole argument.

        Args:
            n_steps (int): Number of steps to run the decision making loop for.
            black_box_function_evaluator (FunctionEvaluator): Function evaluator which
                evaluates the black box functions of interest at supplied points.

        Returns:
            Mapping[str, Dataset]: Dictionary of datasets containing the observations
            made throughout the decision making loop, as well as the initial data
            supplied when initialising the `DecisionMaker`.
        """
        for _ in range(n_steps):
            query_point = self.ask(self.key)

            for post_ask_method in self.post_ask:
                post_ask_method(self, query_point)

            self.key, _ = jr.split(self.key)
            observation_datasets = black_box_function_evaluator(query_point)
            self.tell(observation_datasets, self.key)

            for post_tell_method in self.post_tell:
                post_tell_method(self)

        return self.datasets


@dataclass
class UtilityDrivenDecisionMaker(AbstractDecisionMaker):
    """
    UtilityDrivenDecisionMaker class which handles the core decision making loop in a
    typical model-based decision making setup. In this setup we use surrogate model(s)
    for the function(s) of interest, and define a utility function (often called the
    'acquisition function' in the context of Bayesian optimisation) which characterises
    how useful it would be to query a given point within the search space given the data
    we have observed so far. This can then be used to decide which point(s) to query
    next.

    The decision making loop is split into two key steps, `ask` and `tell`. The `ask`
    step forms a `UtilityFunction` from the current `posteriors` and `datasets` and
    returns the point which maximises it. It also stores the formed utility function
    under the attribute `self.current_utility_function` so that it can be called,
    for instance for plotting, after the `ask` function has been called. The `tell` step
    adds a newly queried point to the `datasets` and updates the `posteriors`.

    This can be run as a typical ask-tell loop, or the `run` method can be used to run
    the decision making loop for a fixed number of steps. Moreover, the `run` method executes
    the functions in `post_ask` and `post_tell` after each ask and tell step
    respectively. This enables the user to add custom functionality, such as the ability
    to plot values of interest during the optimization process.

    Attributes:
        utility_function_builder (AbstractUtilityFunctionBuilder): Object which
                builds utility functions from posteriors and datasets, to decide where
                to query next. In a typical Bayesian optimisation setup the point chosen to
                be queried next is the point which maximizes the utility function.
        utility_maximizer (AbstractUtilityMaximizer): Object which maximizes
            utility functions over the search space.
    """

    utility_function_builder: AbstractUtilityFunctionBuilder
    utility_maximizer: AbstractUtilityMaximizer

    def __post_init__(self):
        super().__post_init__()
        if self.batch_size > 1 and not isinstance(
            self.utility_function_builder, ThompsonSampling
        ):
            raise NotImplementedError(
                "Batch size > 1 currently only supported for Thompson sampling."
            )

    def ask(self, key: KeyArray) -> Float[Array, "B D"]:
        """
        Get updated utility function(s) and return the point(s) which maximises it/them. This
        method also stores the utility function(s) in
        `self.current_utility_functions` so that they can be accessed after the ask
        function has been called. This is useful for non-deterministic utility
        functions, which may differ between calls to `ask` due to the splitting of
        `self.key`.

        Note that in general `SinglePointUtilityFunction`s are only capable of
        generating one point to be queried at each iteration of the decision making loop
        (i.e. `self.batch_size` must be 1). However, Thompson sampling can be used in a
        batched setting by drawing a batch of different samples from the GP posterior.
        This is done by calling `build_utility_function` with different keys
        sequentilly, and optimising each of these individual samples in sequence in
        order to obtain `self.batch_size` points to query next.

        Args:
            key (KeyArray): JAX PRNG key for controlling random state.

        Returns:
            Float[Array, "B D"]: Point(s) to be queried next.
        """
        self.current_utility_functions = []
        maximizers = []
        # We currently only allow Thompson sampling to be run with batch size > 1. More
        # batched utility functions may be added in the future.
        if isinstance(self.utility_function_builder, ThompsonSampling) or (
            (not isinstance(self.utility_function_builder, ThompsonSampling))
            and (self.batch_size == 1)
        ):
            # Draw 'self.batch_size' Thompson samples and optimize each of them in order to
            # obtain 'self.batch_size' points to query next.
            for _ in range(self.batch_size):
                decision_function = (
                    self.utility_function_builder.build_utility_function(
                        self.posteriors, self.datasets, key
                    )
                )
                self.current_utility_functions.append(decision_function)

                _, key = jr.split(key)
                maximizer = self.utility_maximizer.maximize(
                    decision_function, self.search_space, key
                )
                maximizers.append(maximizer)
                _, key = jr.split(key)

            maximizers = jnp.concatenate(maximizers)
            return maximizers
        else:
            raise NotImplementedError(
                "Only Thompson sampling currently supports batch size > 1."
            )
